"""Langevin dynamics class."""

import numpy as np
from numpy.random import standard_normal, randint
import asap3
import ase
import ase.units
from asap3.md.md import ParallelMolDynMixin
from ase.md.md import MolecularDynamics
from ase.md.langevin import Langevin as _Langevin_ASE
import asap3.constraints
from ase.parallel import world, DummyMPI
import sys
import numbers

class ASE_Langevin(_Langevin_ASE, ParallelMolDynMixin):
    def __init__(self, atoms, timestep, temperature=None, friction=None, fixcm=True,
                     *, temperature_K=None, seed=None, **kwargs):
        ParallelMolDynMixin.__init__(self, "Langevin", atoms)
        if seed is not None:
            if 'rng' in kwargs:
                raise ValueError('ASE_Langevin: Cannot specify both seed and rng parameters!')
            if not isinstance(seed, np.random.SeedSequence):
                seed = np.random.SeedSequence(seed)
            kwargs['rng'] = np.random.default_rng(seed.spawn(world.size)[world.rank])
        if ase.__version__ >= '3.25.0':
            _Langevin_ASE.__init__(self, atoms, timestep, temperature=temperature,
                                   friction=friction, temperature_K=temperature_K,
                                   fixcm=fixcm, comm=DummyMPI(), **kwargs)
        else:
            _Langevin_ASE.__init__(self, atoms, timestep, temperature=temperature,
                                   friction=friction, temperature_K=temperature_K,
                                   fixcm=fixcm, communicator=None, **kwargs)

    def run(self, steps):
        self.before_run()
        super().run(steps)
        self.after_run()

    def _get_com_velocity(self, velocity=None):
        """Return the center of mass velocity."""
        if velocity is None:
            velocity = self.v  # Compatibility with older ASE
        if getattr(self.atoms, "parallel", False):
            data = np.zeros(4)
            data[:3] = np.dot(self.masses.flatten(), velocity)
            data[3] = self.masses.sum()
            self.atoms.comm.sum(data)
            return data[:3] / data[3]
        else:
            return np.dot(self.masses.flatten(), velocity) / self.masses.sum()

    temp = property(lambda s: s.get("temp"), lambda s, x: s.set("temp", x))
    fr = property(lambda s: s.get("fr"), lambda s, x: s.set("fr", x))
    masses = property(lambda s: s.get("masses"), lambda s, x: s.set("masses", x))
    c1 = property(lambda s: s.get("c1"), lambda s, x: s.set("c1", x))
    c2 = property(lambda s: s.get("c2"), lambda s, x: s.set("c2", x))
    c3 = property(lambda s: s.get("c3"), lambda s, x: s.set("c3", x))
    c4 = property(lambda s: s.get("c4"), lambda s, x: s.set("c4", x))
    c5 = property(lambda s: s.get("c5"), lambda s, x: s.set("c5", x))
    v = property(lambda s: s.get("v"), lambda s, x: s.set("v", x))
    rnd_pos = property(lambda s: s.get("rnd_pos"), lambda s, x: s.set("rnd_pos", x))
    rnd_vel = property(lambda s: s.get("rnd_vel"), lambda s, x: s.set("rnd_vel", x))

class Langevin_Fast(MolecularDynamics, ParallelMolDynMixin):
    def __init__(self, atoms, timestep, temperature=None, friction=None, fixcm=True,
                     *, temperature_K=None, 
                     trajectory=None, logfile=None, loginterval=1, seed=None):
        temperature = ase.units.kB * self._process_temperature(temperature, temperature_K, 'eV')
        ParallelMolDynMixin.__init__(self, "Langevin", atoms)
        self._uselocaldata = False # Need to store on atoms for serial simul too.
        self.calculator = atoms.calc
        if not atoms.has('momenta'):
            atoms.set_momenta(np.zeros((len(atoms), 3), float))
        self.atoms_have_constraints = len(atoms.constraints)
        if self.atoms_have_constraints:
            if len(atoms.constraints) != 1:
                raise RuntimeError("ASAP Langevin dynamics only support a single constraint.")
            constraint = atoms.constraints[0]
            if not isinstance(constraint, asap3.constraints.FixAtoms):
                raise RuntimeError("ASAP Langevin dynamics only support the ASAP FixAtoms constraint.")
            # Make all constants arrays by making friction an array
            friction = friction * np.ones(len(atoms))
            fixcm = False   # Unneccesary (and incompatible) when FixAtoms constraint used.
        if seed is None:
            seed = randint(1 << 30)
        elif isinstance(seed, np.random.SeedSequence):
            # Convert into a numeric seed for the Asap implementation.
            seed = seed.generate_state(1, dtype=np.uint64)
        assert isinstance(seed, numbers.Integral), "seed must be an int"
        self.asap_md = asap3._asap.Langevin(atoms, self.calculator, timestep,
                                            self.prefix+"sdpos", self.prefix+"sdmom",
                                            self.prefix+"c1", self.prefix+"c2",
                                            fixcm, seed)
        MolecularDynamics.__init__(self, atoms, timestep, trajectory,
                                   logfile, loginterval)
        self.temp = temperature
        self.frict = friction
        self.fixcm = fixcm  # will the center of mass be held fixed?
        self.communicator = None
        self.updatevars()

    def set_temperature(self, temperature=None, *, temperature_K=None):
        self.temp =  ase.units.kB * self._process_temperature(temperature, temperature_K, 'eV')
        self.updatevars()

    def set_friction(self, friction):
        self.frict = friction
        self.updatevars()

    def set_timestep(self, timestep):
        self.dt = timestep
        self.updatevars()

    def updatevars(self):
        dt = self.dt
        # If the friction is an array some other constants must be arrays too.
        self._localfrict = hasattr(self.frict, 'shape')
        lt = self.frict * dt
        masses = self.masses
        sdpos = dt * np.sqrt(self.temp / masses.reshape(-1) *
                             (2.0 / 3.0 - 0.5 * lt) * lt)
        sdpos.shape = (-1, 1)
        sdmom = np.sqrt(self.temp * masses.reshape(-1) * 2.0 * (1.0 - lt) * lt)
        sdmom.shape = (-1, 1)
        pmcor = np.sqrt(3.0) / 2.0 * (1.0 - 0.125 * lt)
        cnst = np.sqrt((1.0 - pmcor) * (1.0 + pmcor))

        act0 = 1.0 - lt + 0.5 * lt * lt
        act1 = (1.0 - 0.5 * lt + (1.0 / 6.0) * lt * lt)
        act2 = 0.5 - (1.0 / 6.0) * lt + (1.0 / 24.0) * lt * lt
        c1 = act1 * dt / masses.reshape(-1)
        c1.shape = (-1, 1)
        c2 = act2 * dt * dt / masses.reshape(-1)
        c2.shape = (-1, 1)
        c3 = (act1 - act2) * dt
        c4 = act2 * dt
        del act1, act2
        if self._localfrict:
            # If the friction is an array, so are these
            act0.shape = (-1, 1)
            c3.shape = (-1, 1)
            c4.shape = (-1, 1)
            pmcor.shape = (-1, 1)
            cnst.shape = (-1, 1)
        self.sdpos = sdpos
        self.sdmom = sdmom
        self.c1 = c1
        self.c2 = c2
        self.act0 = act0
        self.c3 = c3
        self.c4 = c4
        self.pmcor = pmcor
        self.cnst = cnst
        # Also works in parallel Asap:
        self.natoms = self.atoms.get_global_number_of_atoms() #GLOBAL number of atoms
        if len(self.atoms.constraints) == 1:
            # Process the FixAtoms constraint
            constr = self.atoms.constraints[0].index
            self.sdpos[constr] = 0.0
            self.sdmom[constr] = 0.0
            self.c1[constr] = 0.0
            self.c2[constr] = 0.0
            self.c3[constr] = 0.0
            self.c4[constr] = 0.0
            self.act0[constr] = 0.0
        if self._localfrict:
            self.asap_md.set_vector_constants(self.prefix+"act0", self.prefix+"c3",
                                              self.prefix+"c4", self.prefix+"pmcor",
                                              self.prefix+"cnst")
        else:
            self.asap_md.set_scalar_constants(self.act0, self.c3, self.c4,
                                              self.pmcor, self.cnst)

    def run(self, steps):
        assert(self.calculator is self.atoms.calc)
        if self.atoms.constraints or self.atoms_have_constraints:
            # Constraints may have been added, removed or changed
            self.updatevars()
        self.before_run()
        self.asap_md.run(steps, self.observers, self)
        self.after_run()
        self.atoms_have_constraints = len(self.atoms.constraints)

    def get_random(self, gaussian):
        return self.asap_md.get_random(gaussian)

    # Properties are not inherited, need to repeat them
    sdpos = property(lambda s: s.get("sdpos"), lambda s, x: s.set("sdpos", x))
    sdmom = property(lambda s: s.get("sdmom"), lambda s, x: s.set("sdmom", x))
    c1 = property(lambda s: s.get("c1"), lambda s, x: s.set("c1", x))
    c2 = property(lambda s: s.get("c2"), lambda s, x: s.set("c2", x))
    act0 = property(lambda s: s.get("act0"), lambda s, x: s.set("act0", x))
    c3 = property(lambda s: s.get("c3"), lambda s, x: s.set("c3", x))
    c4 = property(lambda s: s.get("c4"), lambda s, x: s.set("c4", x))
    pmcor = property(lambda s: s.get("pmcor"), lambda s, x: s.set("pmcor", x))
    cnst = property(lambda s: s.get("cnst"), lambda s, x: s.set("cnst", x))
    masses = property(lambda s: s.get("masses"), lambda s, x: s.set("masses", x))
    frict = property(lambda s: s.get("frict"), lambda s, x: s.set("frict", x))

def Langevin(atoms, *args, **kwargs):
    if (isinstance(atoms, ase.Atoms)
        and asap3.constraints.check_asap_constraints(atoms)
        and not kwargs.pop('forceASE', False)
       ):
        # Nothing prevents Asap optimization
        if world.rank == 0:
            sys.stderr.write("Using Asap-optimized C++-Langevin algorithm\n")
        return Langevin_Fast(atoms, *args, **kwargs)
    else:
        if world.rank == 0:
            sys.stderr.write("Using ASE-based Langevin algorithm\n")
        return ASE_Langevin(atoms, *args, **kwargs)



import sys
import time
import numpy as np
import os
import numbers

from ase.parallel import world
from ase.utils import basestring

from asap3.Tools.ParameterOptimization.Optimization import ParameterOptimization

def now():
    "Time as a string."
    return time.strftime("%H:%M:%S")

class ParameterSearch(ParameterOptimization):
    """Parallel parameter optimization."""
    def __init__(self, elements, calc_class, initparam, varparam, quantities,
                 latticeconstants, calc_args=None, debug=False):
        ParameterOptimization.__init__(self, elements, calc_class, initparam,
                                       varparam, quantities, latticeconstants,
                                       calc_args, debug)
        if debug:
            sys.stdout = open(f"process-{world.rank}.stdout", "w", 1)

    def random_parameters(self, parameters, std):
        change = np.random.normal(1.0, std, len(parameters))
        return (np.array(parameters) * change).tolist()

    def run(self, std, temp, maxstep=100, xtol=1e-2, ftol=5e-2, delta=5e-3,
            ptol=1e-2, dat=None, log=None, err=None):
        """Run the optimization.
        
        First parameter: Gauss width for MC part
        Second parameter: MC Temperature
        Third parameter: run time, in h or m.  Or the number of steps if integer.
        xtol and ftol are simplex convergence criterial.  
        delta is the initial simplex size, relatively.
        """
        self.ptol = ptol
        self.nsteps = 0
        self.numberlist = []
        self.errorlist = []
        self.countlist = []
        self.fitparlist = []
        self.optparlist = []
        self.optvallist = []

        self.maxstep = None
        self.endtime = None
        self.done(maxstep)

        if log != None:
            s = log.rfind('.')
            log = open(log[:s] + '-' + str(world.rank) + log[s:], 'a')

        if err != None:
            s = err.rfind('.')
            err = open(err[:s] + '-' + str(world.rank) + err[s:], 'w')

        if world.rank == 0:
            done = np.zeros(world.size)
            request_s = [0] * world.size
            request_r = [0] * world.size
            result = [0] * world.size
        else:
            done = False
            request_s = None
            request_r = None
            result = np.empty(1)

        accepts = 0
        old_error = 1e6
        old_optpar = None
        while not self.done():
            # Make a random set of parameters close to the
            if old_optpar == None:
                initparam = self.parameters_dict2list()
                if world.rank == 0:
                    fitpar = initparam
                else:
                    fitpar = self.random_parameters(initparam, std * 2.0)
            else:
                fitpar = self.random_parameters(old_optpar, std)

            # Run optimization
            (error, optpar, optval) = self.fit(fitpar, xtol, ftol, delta,
                                               log, err, True)

            self.nsteps += 1
            if error < old_error or temp < 1e-6:
                prop = 1.0
            else:
                derr = (error - old_error) / old_error
                prop = np.exp(-derr/temp)

            if prop > np.random.uniform():
                accepts += 1
                old_error = error
                old_optpar = optpar[:]
                self.save_opt(error, fitpar, optpar, optval)
         
        # Gather results and Write output
        sys.stderr.write("Task %i [%s]: Finished optimizing, writing results.\n"
                         % (world.rank, now()))
        self.write_local(dat, std, temp, accepts)
        if dat != None:
            sys.stderr.write("Task %i [%s]: Communicating.\n" % (world.rank, now()))
            a = np.zeros(21)
            alen = min(len(self.errorlist), 21)
            a[:alen] = np.array(self.errorlist)[:alen]
            if world.rank == 0:
                errors = np.zeros(world.size * len(a))
                tasks = np.arange(world.size * len(a)) / 21
                if world.size > 1:
                    world.gather(a, 0, errors)
                else:
                    errors = a
                tasks = tasks[errors > 1e-20]
                errors = errors[errors > 1e-20]
                tasks = tasks[errors.argsort()]
                errors = errors[errors.argsort()]
                f = open(dat, 'a')
                f.write('\nParameter search at ' + time.asctime() + '\n')
                f.write('--------------------------------------------\n')
                f.write('Rank  Task  Error function\n')
                for i, (n, e) in enumerate(zip(tasks, errors)):
                    f.write('%-4i  %-4i  %.5e\n' % (i, n, e))
                f.close()
            else:
                world.gather(a, 0)
        sys.stderr.write("Task %i [%s]: Done.\n" % (world.rank, now()))


    def done(self, maxstep=None, hard=False):
        if maxstep != None:
            if isinstance(maxstep, numbers.Integral):
                self.maxstep = maxstep
            elif isinstance(maxstep, basestring):
                multiplier = {'h': 3600, 'm': 60}
                elapse = float(maxstep[:-1]) * multiplier[maxstep[-1]]
                self.endtime = time.time() + elapse
            else:
                raise ValueError('Cannot understand the given maxstep.')
        else:
            if self.maxstep == None:
                return time.time() >= self.endtime
            else:
                if hard:
                    return self.nsteps > self.maxstep * 1.2
                else:
                    return self.nsteps >= self.maxstep

    def write_local(self, dat, std, temp, accepts):
        s = dat.rfind('.')
        dat = open(dat[:s] + '-' + str(world.rank) + dat[s:], 'a')

        T = time.localtime()
        dat.write("Parameter search %02i-%02i-%i " % (T[2], T[1], T[0]) +
                  "at %02i:%02i:%02i\n" % (T[3], T[4], T[5]) +
                  "-" * 90 + "\n" +
                  "%i optimizations has been made and " % (self.nsteps,) +
                  "%i have been accepted.\n" % (accepts,) +
                  "Below the initial and the 20 best are listed.\n\n" +
                  f"Gauss width: {std:.3f}\nTemperature: {temp:.3f}\n\n")

        for num, err, count, fitpar, optpar, optval in zip(self.numberlist,
                                                           self.errorlist,
                                                           self.countlist,
                                                           self.fitparlist,
                                                           self.optparlist,
                                                           self.optvallist):
            dat.write("Optimization no. %i with " % (num,) +
                      "error function %.5e (%i):\n" % (err, count))
            self.write_result(fitpar, optpar, optval, dat)
        dat.close()

    def save_opt(self, error, fitpar, optpar, optval):
        # Sort and save the 20 best results
        length = len(self.errorlist)
        if length == 0:
            self.numberlist.append(self.nsteps)
            self.errorlist.append(error)
            self.countlist.append(1)
            self.fitparlist.append(fitpar)
            self.optparlist.append(optpar)
            self.optvallist.append(optval)
        else:
            new_min = -1
            new_rank = -1
            for i, err in enumerate(self.errorlist):
                op1 = np.array(optpar)
                op2 = np.array(self.optparlist[i])

                #ov1 = np.array(optval)
                #ov2 = np.array(self.optvallist[i])

                #ov_dev_max = np.max(2 * np.abs(ov1 - ov2) / np.abs(ov1 + ov2))
                op_dev_max = np.max(2 * np.abs(op1 - op2) / np.abs(op1 + op2))

                if op_dev_max < self.ptol and new_min < 0 and i > 0: #ov_dev_max < 2e-2 and
                    new_min = i

                if error < err and new_rank < 0 and i > 0:
                    new_rank = i

            #print new_min, new_rank, length
            if new_min < 0:
                # New minimum found
                if new_rank > 0:
                    # Better error function
                    self.numberlist.insert(new_rank, self.nsteps)
                    self.errorlist.insert(new_rank, error)
                    self.countlist.insert(new_rank, 1)
                    self.fitparlist.insert(new_rank, fitpar)
                    self.optparlist.insert(new_rank, optpar)
                    self.optvallist.insert(new_rank, optval)
                    if length >= 21:
                        del self.numberlist[-1]
                        del self.errorlist[-1]
                        del self.countlist[-1]
                        del self.fitparlist[-1]
                        del self.optparlist[-1]
                        del self.optvallist[-1]
                else:
                    # Worst error function
                    if length < 21:
                        self.numberlist.append(self.nsteps)
                        self.errorlist.append(error)
                        self.countlist.append(1)
                        self.fitparlist.append(fitpar)
                        self.optparlist.append(optpar)
                        self.optvallist.append(optval)
            else:
                # Old minimum found
                if  new_rank > 0 and new_rank <= new_min:
                    # Better error function
                    del self.numberlist[new_min]
                    del self.errorlist[new_min]
                    count = self.countlist.pop(new_min)
                    del self.fitparlist[new_min]
                    del self.optparlist[new_min]
                    del self.optvallist[new_min]
                    self.numberlist.insert(new_rank, self.nsteps)
                    self.errorlist.insert(new_rank, error)
                    self.countlist.insert(new_rank, count + 1)
                    self.fitparlist.insert(new_rank, fitpar)
                    self.optparlist.insert(new_rank, optpar)
                    self.optvallist.insert(new_rank, optval)
                else:
                    self.countlist[new_min] += 1


if __name__ == '__main__':
    from asap3.Tools.ParameterOptimization.EMT import EMT2011Fit

    # Parameter names: ['eta2', 'lambda', 'kappa', 'E0', 'V0', 'S0', 'n0']
    initparam = {('Pt','Pt'): [3.4242, 4.1423, 5.9432, -5.85, 4.067, 1.5346, 0.05412]}
    varparam = {('Pt','Pt'): [True, True, True, True, True, True, False]}

    quantities = [('lattice_constant_a', 'fcc', 'Pt', 3.92,      0.001),
                  #('lattice_constant_a', 'hcp', 'Pt', 2.77,      0.05),
                  #('lattice_ratio_ca', 'hcp', 'Pt', 4.78 / 2.77, 0.05),
                  ('bulk_modulus', 'fcc', 'Pt', 278.3,           0.01),
                  #('elastic_constant_C11', 'fcc', 'Pt', 346.7,   0.1),
                  #('elastic_constant_C12', 'fcc', 'Pt', 250.7,  0.1),
                  #('elastic_constant_C44', 'fcc', 'Pt', 76.5,    0.02),
                  ('cohesive_energy', 'fcc', 'Pt', 5.84,         0.01),
                  #('phase_energy', 'fcc-hcp', 'Pt', -0.05,       0.02),
                  #('surface_energy', 'fcc111', 'Pt', 0.631,      0.02),
                  #('surface_ratio', 'fcc111-fcc100', 'Pt', 0.631 / 0.892, 0.01),
                  ]

    latticeconstants = [['fcc', 'Pt', 3.9],
                        ['hcp', 'Pt', (2.8, 4.8)]]

    opt = ParameterSearch(['Pt'], EMT2011Fit, initparam, varparam, quantities,
                          latticeconstants, ('kappa',), True)
    opt.run(0.01, 0.0, '0.1m', dat='SearchParallel.dat', log='SearchParallel.log',
            err='SearchParallel.err')





#!/usr/bin/env python
"""Tests that lattices are set up correctly by Asap

Description: Part of the Asap test suite.  Tests that FCC, BCC and HCP lattices
    are set up correctly.
"""

import sys
from numpy import *
from asap3 import *
from asap3.optimize.mdmin import MDMin
from asap3.analysis import CNA, CoordinationNumbers
from ase.lattice.cubic import *
from ase.lattice.orthorhombic import BodyCenteredOrthorhombic,\
     FaceCenteredOrthorhombic
from ase.lattice.hexagonal import HexagonalClosedPacked
from ase.lattice.triclinic import Triclinic
from ase.lattice.monoclinic import BaseCenteredMonoclinic
from asap3.test.pytest_markers import ReportTest, serial

import pytest

def checkFCC(atoms, n, name):
    print("Test '%s': %d atoms" % (name, len(atoms)))
    ReportTest((f"Number of atoms ({name})"), len(atoms), n, 0)
    atoms.calc = EMT()
    cn = CoordinationNumbers(atoms)
    ReportTest((f"Coordination number is 12 ({name})"),
               sum(equal(cn, 12)), len(atoms), 0)
    cna = CNA(atoms)
    ReportTest((f"CNA says FCC ({name})"),
               sum(equal(cna, 0)), len(atoms), 0)
    epot = atoms.get_potential_energy()/len(atoms)
    ReportTest((f"Potential energy ({name})"), epot, 0.0, 1e-3)

def checkBCC(atoms, n, name):
    print("Test '%s': %d atoms" % (name, len(atoms)))
    ReportTest((f"Number of atoms ({name})"), len(atoms), n, 0)
    cn = CoordinationNumbers(atoms, 2.94)
    ReportTest((f"Coordination number is 8 ({name})"),
               sum(equal(cn, 8)), len(atoms), 0)
    cna = CNA(atoms)
    ReportTest((f"CNA says neither FCC nor HCP ({name})"),
               sum(equal(cna, 2)), len(atoms), 0)

def checkHCP(atoms, n, name):
    print("Test '%s': %d atoms" % (name, len(atoms)))
    ReportTest((f"Number of atoms ({name})"), len(atoms), n, 0)
    atoms.calc = EMT()
    cn = CoordinationNumbers(atoms)
    ReportTest((f"Coordination number is 12 ({name})"),
               sum(equal(cn, 12)), len(atoms), 0)
    cna = CNA(atoms)
    ReportTest((f"CNA says HCP ({name})"),
               sum(equal(cna, 1)), len(atoms), 0)
    epot1 = atoms.get_potential_energy()/len(atoms)
    print("Potential energy:", epot1, "eV/atom")
    r1 = atoms.get_positions()
    dyn = MDMin(atoms, dt=2*units.fs)
    dyn.run(fmax=0.001, steps=50)
    epot2 = atoms.get_potential_energy()/len(atoms)
    ReportTest((f"Change in potential energy ({name})"), epot2-epot1,
               0.0, 1e-6)


@serial
@pytest.mark.core
def test_lattice_fcc():
    a = FaceCenteredCubic(size=(10,10,10), symbol="Cu", pbc=(1,1,1))
    checkFCC(a, 4000, "FCC 1")

    dirs = [[1,5,2], [-3,-1,4]]
    dirs.append(cross(dirs[0], dirs[1]))
    a = FaceCenteredCubic(directions=dirs, size=(2,2,1), symbol="Cu",
                        pbc=(1,1,1))
    checkFCC(a, 1560, "FCC 2")

    directions = [[5,1,3], [3,5,1], [2,2,-7]]
    a = FaceCenteredCubic(directions=directions, size=(2,2,2), symbol="Cu")
    checkFCC(a, 5568, "FCC 4")

    directions = [[5,1,3], [3,5,1], [-2,-2,7]]
    a = FaceCenteredCubic(directions=directions, size=(2,2,2), symbol="Cu")
    checkFCC(a, 5568, "FCC 5")

    a = FaceCenteredCubic(directions=[[1,0,0], [0,1,0], [0,0,1]], size=(10,10,10),
                        symbol=29)
    checkFCC(a, 4000, "FCC 7")

    a = FaceCenteredCubic(directions=[[1,0,0], [0,1,0], [0,0,-1]], size=(10,10,10),
                        symbol="Cu")
    checkFCC(a, 4000, "FCC 8")

    a = FaceCenteredCubic(directions=[[1,0,0], [0,-1,0], [0,0,1]], size=(10,10,10),
                        symbol="Cu")
    checkFCC(a, 4000, "FCC 9")

    a = FaceCenteredCubic(directions=[[-1,0,0], [0,1,0], [0,0,1]], size=(10,10,10),
                        symbol="Cu")
    checkFCC(a, 4000, "FCC 10")

    a = FaceCenteredCubic(directions=[[-1,0,0], [0,-1,0], [0,0,-1]], size=(10,10,10),
                        symbol="Cu")
    checkFCC(a, 4000, "FCC 11")

@serial
@pytest.mark.core
def test_lattice_bcc():
    a = BodyCenteredCubic(directions=[[1,0,0], [0,1,0], [0,0,1]], size=(10,10,10),
                        symbol="Mo")
    checkBCC(a, 2000, "BCC 1")

    a = BodyCenteredCubic(directions=[[8,5,2], [-1,0,4], [20,-34,5]], size=(2,1,1),
                        symbol="Mo")
    checkBCC(a, 12648/2, "BCC 2")

    a = BodyCenteredCubic(directions=[[8,5,3], [-1,1,1], [3, 5, -2]], size=(3,4,3),
                        symbol="Mo")
    checkBCC(a, 75*3*4*3, "BCC 3")

    a = BodyCenteredCubic(directions=[[1,0,0], [0,1,0], [0,0,1]], size=(10,10,10),
                        symbol="Mo")
    checkBCC(a, 2000, "BCC 4")

    a = BodyCenteredCubic(directions=[[-1,0,0], [0,1,0], [0,0,1]], size=(10,10,10),
                        symbol="Mo")
    checkBCC(a, 2000, "BCC 5")

@serial
@pytest.mark.core
def test_lattice_bcorthorhombic():
    a = BodyCenteredOrthorhombic(directions=[[1,0,0], [0,1,0], [0,0,1]],
                                size=(10,10,10),
                                symbol="Mo", latticeconstant=(3.15, 3.15, 3.15))
    checkBCC(a, 2000, "BC Orthorhombic/BCC")

    a = BodyCenteredOrthorhombic(directions=[[1,0,0], [0,1,0], [0,0,1]],
                                size=(10,10,10),
                                symbol="Cu", latticeconstant=(3.61/sqrt(2),
                                                            3.61/sqrt(2), 3.61))
    checkFCC(a, 2000, "BC Orthorhombic/FCC")

@serial
@pytest.mark.core
def test_lattice_fcorthorhombic():
    a = FaceCenteredOrthorhombic(directions=[[1,0,0], [0,1,0], [0,0,1]],
                                size=(10,10,10),
                                symbol="Cu", latticeconstant=(3.61, 3.61, 3.61))
    checkFCC(a, 4000, "FC Orthorhombic/FCC")

@serial
@pytest.mark.core
def test_lattice_triclinic():
    a0 = 3.61 / sqrt(2.0)
    a = Triclinic(directions=[[2,1,0],[-2,3,0],[0,1,4]], size=(3,3,3), symbol="Cu",
                latticeconstant=(a0,a0,a0,60,60,60))
    checkFCC(a, 32*3*3*3, "Triclinic/FCC 1")
                
    a = Triclinic(directions=[[-1,1,1],[1,-1,1],[1,1,-1]], size=(4,4,4), symbol="Cu",
                latticeconstant=(a0,a0,a0,60,60,60))
    checkFCC(a, 4*4*4*4, "Triclinic/FCC 2")

    a = Triclinic(directions=[[2,1,0],[-2,3,0],[0,1,4]], size=(3,3,3), symbol="Cu",
                latticeconstant=(3.61,a0,a0,60,45,45))
    checkFCC(a, 32*3*3*3, "Triclinic/FCC 3")

    a = BaseCenteredMonoclinic(directions=[[2,1,0],[-2,3,0],[0,1,4]], size=(3,3,3),
                            symbol="Cu", latticeconstant=(3.61,3.61,a0,45))
    checkFCC(a, 64*3*3*3, "Monoclinic/FCC")


    a0 = 3.15 * sqrt(3)/2
    a = Triclinic(directions=[[2,1,0],[-2,3,0],[0,1,4]], size=(3,3,3), symbol="Mo",
                latticeconstant=(3.15,3.15,a0,54.736,54.736,90))
    checkBCC(a, 32*3*3*3, "Triclinic/BCC")

@serial
@pytest.mark.core
def test_lattice_hcp():
    a = HexagonalClosedPacked(directions=[[2,-1,-1,0], [0,1,-1,0], [0,0,0,1]],
                            size=(10,5,5),
                            symbol="Cu", 
                            latticeconstant={'a': 3.61/sqrt(2),
                                            'c/a':2*sqrt(2)/sqrt(3)})
    checkHCP(a, 1000, "HCP 1")




from asap3 import *
from asap3.mpi import world
from asap3.io import Trajectory
from ase.io import write
from ase.build import bulk
from asap3.test.pytest_markers import ReportTest, parallel

import pytest

@parallel
def test_autodistribution(in_tmp_dir):
    'Tests the auto keyword for cpulayout when loading from a Trajectory'
    if world.rank == 0:
        atoms1 = bulk("Cu", cubic=True).repeat((20,20,20))
        atoms1.set_pbc = False
        atoms1.center(vacuum=5.)
        ReportTest("Number of atoms", len(atoms1), 32000, 0)
        atoms1.calc = EMT()
        e0 = atoms1.get_potential_energy()
        write("serialatoms.traj", atoms1, parallel=False)
    else:
        atoms1 = None

    world.barrier()

    atoms = Trajectory("serialatoms.traj").get_atoms(-1, 'auto')
    atoms.calc = EMT()
    e = atoms.get_potential_energy()
    world.barrier()   # Protects cleanup
    if world.rank == 0:
        os.unlink('serialatoms.traj')
        ReportTest("Energy of saved atoms", e, e0, 1e-3)

    world.barrier()



from ase.data import atomic_numbers, chemical_symbols, reference_states
from ase.units import *

from . import fcc
from . import au

lattice = {'fcc': fcc.data,
          }

element = {79: au.data, #Au
          }



import numpy as np
import numbers

def check_parameter_array(numelements, name, a):
    if numelements > 1:
        ao = np.array(a)

        shape = (numelements, numelements)
        assert(ao.shape == shape)

        an = np.zeros(shape)

        for i in range(numelements):
            for j in range(i, numelements):
                an[i,j] = ao[i,j]
                an[j,i] = ao[i,j]

    else:
        if isinstance(a, numbers.Real):
            an = np.array([[a]])
        else:
            an = a.copy()

    return an

def _ChkLJarray(data, n, var, pot="LennardJones"):
    "Check the parameters of the LennardJones factory function."
    # Is data a scalar or an array?
    try:
        l = len(data)
    except TypeError:
        if n == 1:
            return [data]
        else:
            raise TypeError("%s: %s parameter cannot be a scalar (%d elements)."
                            % (pot, var, n))
    # Get the shape
    try:
        shp = data.shape
    except AttributeError:
        data = np.array(data)
        shp = data.shape
    if shp != (n,n) and shp != (n*n,):
        raise TypeError("%s: %s parameter should be a %dx%d array"
                        % (pot, var, n, n))
    if shp == (n,n):
        for i in range(n):
            for j in range(i+1,n):
                if data[i,j] != 0 and data[i,j] != data[j,i]:
                    raise ValueError(f"{pot}: {var} matrix must be a symmetric or a lower triangular matrix")
    return data.flat[:]




from asap3 import Atoms, MakeParallelAtoms, LennardJones, units
from ase.md.verlet import VelocityVerlet
from ase.lattice.cubic import FaceCenteredCubic
from asap3.io.trajectory import Trajectory
from numpy import sqrt, sin, arange
from asap3.mpi import world
from asap3.test.pytest_markers import ReportTest, serial

import pytest


elements = [29]
epsilon  = [0.15]
sigma    = [2.7]

@pytest.fixture
def initial():
    if world.rank == 0:
        initial = FaceCenteredCubic(directions=((1,0,0),(0,1,0),(0,0,1)),
                                    size=(40,40,40), symbol="Cu",
                                    latticeconstant=1.09*sigma[0]*1.41,
                                    pbc=(1,1,0))
        momenta = sqrt(2*63.5 * units.kB * 400) * sin(arange(3*len(initial)))
        momenta.shape = (-1,3)
        initial.set_momenta(momenta)
        print("Number of atoms:", len(initial))
    else:
        initial = None
    return initial

@pytest.mark.slow
def test_LennardJones(initial, multicpulayout):

    for layout in multicpulayout:
        if layout:
            print("Test with layout "+str(layout))
            atoms = MakeParallelAtoms(initial, layout)
            natoms = atoms.get_global_number_of_atoms()
        else:
            print("Serial test")
            atoms = Atoms(initial)
            natoms = len(atoms)
        print("Number of atoms:", natoms)
        temp = atoms.get_kinetic_energy() / (1.5*units.kB*natoms)
        print("Temp:", temp, "K")
        ReportTest("Initial temperature", temp, 400.0, 1.0)
        atoms.calc = LennardJones(elements, epsilon, sigma, -1.0, True)

        epot = atoms.get_potential_energy()
        print("Potential energy:", epot)
        ReportTest("Initial potential energy", epot, -301358.3, 0.5)
        etot = epot + atoms.get_kinetic_energy()


        dyn = VelocityVerlet(atoms, 3*units.fs)
        if 0:
            traj = Trajectory("llj.traj", "w", atoms)
            dyn.attach(traj, interval=2)
                    
        etot2 = None
        for i in range(5):
            dyn.run(15)
            newetot = atoms.get_potential_energy()+ atoms.get_kinetic_energy()
            print("Total energy:", newetot)
            temp = atoms.get_kinetic_energy() / (1.5*units.kB*natoms)
            print("Temp:", temp, "K")
            if etot2 == None:
                ReportTest("Total energy (first step)", newetot, etot, 40.0)
                etot2=newetot
            else:
                ReportTest(("Total energy (step %d)" % (i+1,)),
                        newetot, etot2, 20.0)
        print(" *** This test completed ***")




import sys
import time
import signal
import numpy as np
import numbers

from asap3.Tools.ParameterOptimization.ScipyFmin import fmin, fmin_bfgs
from asap3.Tools.ElasticConstants import ElasticConstants, minimize_unit_cell
from asap3.Tools.SurfaceEnergies import SurfaceEnergy
from asap3.Tools.AtomicEnergies import CohesiveEnergy, VacancyEnergy

from ase import Atoms
#from ase.io import read
from ase.io.trajectory import Trajectory
from ase.data import reference_states, atomic_numbers, chemical_symbols
from ase.units import kJ, GPa
from ase.optimize import BFGS, QuasiNewton
from ase.filters import StrainFilter
from ase.lattice.cubic import BodyCenteredCubic, FaceCenteredCubic
from ase.lattice.hexagonal import HexagonalClosedPacked
from ase.build import fcc100, fcc110, fcc111, hcp0001
from ase.lattice.compounds import B2, L1_2, L1_0
from ase.lattice.spacegroup import crystal
from ase.cluster.octahedron import Octahedron
from ase.utils import basestring

class QuantityCalc:
    properties = {'fcc_size': (1, 1, 1), # Size of the FCC and HCP unit cell
                  'bcc_size': (1, 1, 1),
                  'hcp_size': (1, 1, 1),
                  'l12_size': (1, 1, 1),
                  'l10_size': (1, 1, 1),
                  'l10c_size': (1, 1, 1),
                  'B2_size': (1, 1, 1),
                  'A5Bhex_size': (1, 1, 1),
                  'fcc100_size': (1, 1), # FCC and HCP surface size
                  'fcc111_size': (1, 2),
                  'hcp0001_size': (1, 2),
                  'surface_layers': [7, 9, 11, 13],
                  'sf_repeat' : (1,1,2),
                  }

    def __init__(self, latticeconstants, calc=None, timer=None, debug=False):
        self.debug = debug
        self.values = {}
        self.oldparams = None

        self.set_calculator(calc)

        if timer != None:
            self.timer = timer
        else:
            self.timer = None

        # Interpet lattice constants
        self.latticeconstants = {}
        for structure, elements, lc in latticeconstants:
            if isinstance(elements, (basestring, numbers.Integral)):
                elements = tuple([elements])
            else:
                elements = tuple(elements)
            self.latticeconstants[(structure, elements)] = lc

    def set_calculator(self, calc=None):
        if calc != None and not (hasattr(calc, 'get_potential_energy')
                                 or callable(calc)):
            raise ValueError('The calculator is not valid')
        self.calc = calc
        self.reset_values()
        
    def reset_values(self):
        """Reset all the calculated values that may have changed"""
        fullreset = (self.calc == None or self.oldparams == None)
        params = None
        if isinstance(self.calc, type):
            print("Full reset performed - calculator is a class.")
            fullreset = True
        else:
            try:
                params = self.calc.get_extra('parameters')
            except (ValueError, AttributeError):
                fullreset = True

        if fullreset:
            self.values = {}
            print("Full reset performed.", file=sys.stderr)
        else:
            # Find which parameters have changed
            nbefore = len(self.values)
            changed = []
            for k,v in params.items():
                assert isinstance(k, numbers.Integral)
                kstr = chemical_symbols[k]
                assert isinstance(kstr, basestring)
                if k not in self.oldparams or v != self.oldparams[k]:
                    # Delete any calculated value that includes this element
                    changed.append(k)
                    for prop in list(self.values.keys()):  # NOT iterkeys !
                        if kstr in prop[2]:
                            del self.values[prop]
            #print >>sys.stderr, ("Resetting %d values of %d: %s" %
            #                     (nbefore - len(self.values), nbefore, str(changed)))
        self.oldparams = params
            
    ### Get quantity functions ###
    def get_quantity(self, name, structure, elements):
        if self.debug:
            print(f"get_quantity({name}, {structure}, {str(elements)})   time: {time.asctime()}")
        func = getattr(self, 'get_' + name)
        return func(structure, elements)

    def get_cohesive_energy(self, structure, elements):
        return self.get_value('cohesive_energy', structure, elements,
                              self.calc_cohesive_energy)

    def get_vacancy_energy(self, structure, elements):
        return self.get_value('vacancy_energy', structure, elements,
                              self.calc_vacancy_energy)

    def get_phase_energy(self, structure, elements):
        return self.get_value('phase_energy', structure, elements,
                              self.calc_phase_energy)

    def get_lattice_constant_a(self, structure, elements):
        return self.get_value('lattice_constant_a', structure, elements,
                              self.calc_lattice_constants)

    def get_lattice_constant_c(self, structure, elements):
        return self.get_value('lattice_constant_c', structure, elements,
                              self.calc_lattice_constants)

    def get_lattice_ratio_ca(self, structure, elements):
        return self.get_value('lattice_ratio_ca', structure, elements,
                              self.calc_lattice_ratio_ca)

    def get_volume_per_atom(self, structure, elements):
        return self.get_value('volume_per_atom', structure, elements,
                              self.calc_volume_per_atom)

    def get_bulk_modulus(self, structure, elements):
        return self.get_value('bulk_modulus', structure, elements,
                              self.calc_elastic_constants)

    def get_elastic_anisotropy(self, structure, elements):
        return self.get_value('elastic_anisotropy', structure, elements,
                              self.calc_elastic_anisotropy)

    def get_elastic_constant_C11(self, structure, elements):
        return self.get_value('elastic_constant_C11', structure, elements,
                              self.calc_elastic_constants)

    def get_elastic_constant_C12(self, structure, elements):
        return self.get_value('elastic_constant_C12', structure, elements,
                              self.calc_elastic_constants)

    def get_elastic_constant_C13(self, structure, elements):
        return self.get_value('elastic_constant_C13', structure, elements,
                              self.calc_elastic_constants)

    def get_elastic_constant_C33(self, structure, elements):
        return self.get_value('elastic_constant_C33', structure, elements,
                              self.calc_elastic_constants)

    def get_elastic_constant_C44(self, structure, elements):
        return self.get_value('elastic_constant_C44', structure, elements,
                              self.calc_elastic_constants)

    def get_surface_energy(self, structure, elements):
        return self.get_value('surface_energy', structure, elements,
                              self.calc_surface_energy)

    def get_surface_ratio(self, structure, elements):
        return self.get_value('surface_ratio', structure, elements,
                              self.calc_surface_ratio)

    def get_heat_of_formation(self, structure, elements):
        return self.get_value('heat_of_formation', structure, elements,
                              self.calc_heat_of_formation)

    #def get_heat_of_sulution(self, structure, elements):
    #    return self.get_value('heat_of_solution', structure, elements,
    #                          self.calc_heat_of_solution)

    def get_impurity_energy(self, structure, elements):
        return self.get_value('impurity_energy', structure, elements,
                              self.calc_impurity_energy)

    def get_impurity_ratio(self, structure, elements):
        return self.get_value('impurity_ratio', structure, elements,
                              self.calc_impurity_ratio)

    def get_cutoff_energy(self, structure, elements):
        return self.get_value('cutoff_energy', structure, elements,
                              self.calc_cutoff_energy)

    def get_scaling_energy(self, structure, elements):
        return self.get_value('scaling_energy', structure, elements,
                              self.calc_scaling_energy)

    def get_force_match(self, structure, elements):
        return self.get_value('force_match', structure, elements,
                              self.calc_force_match)
        
    def get_stacking_fault(self, structure, elements):
        return self.get_value('stacking_fault', structure, elements,
                              self.calc_stacking_fault)
        
    ### Get and set function ###
    def get_value(self, name, structure, elements, function):
        if isinstance(elements, basestring):
            elements = (elements,)
        key = (name, structure, elements)
        if key not in self.values:
            if self.calc != None:
                if self.timer != None and self.debug:
                    self.timer.start(function.__name__)
                try:
                    function(structure, elements)
                except:
                    if self.debug:
                        sys.stderr.write("Calculation Error..." +
                                         f"\n Name: {name}" + 
                                         f"\n Structure: {structure}" +
                                         f"\n Elements: {elements}\n\n")
                        sys.stderr.flush()
                    raise
                finally:
                    if self.timer != None and self.debug:
                        self.timer.stop(function.__name__)
            else:
                raise RuntimeError('No calculator set')

        return self.values[key]

    def set_value(self, name, structure, elements, value):
        key = (name, structure, elements)
        self.values[key] = value

    ### Calculation functions ###
    def calc_cohesive_energy(self, structure, elements):
        if structure not in ['fcc', 'hcp', 'bcc']:
            raise ValueError(f"Cannot calculate cohesive energy for '{structure}'")

        if self.debug:
            print("Calculating cohesive energy...")
            print(40 * "*")
            print(f"Structure: {structure}")
            print(f"Elements: {elements}\n")

        e_full = self.get_value('energy', structure, elements,
                                 self.calc_lattice_constants)

        atoms = Atoms([elements[0]], [(0.0, 0.0, 0.0)])
        atoms.center(vacuum=1.0)
        atoms.set_calculator(self.get_calc())
        e_reduced = atoms.get_potential_energy()

        e = e_reduced - e_full

        self.values[('cohesive_energy', structure, elements)] = e

        if self.debug:
            print(f"Cohesive energy: {e:.5f} eV/atom")
            print(40 * "-" + "\n")

    def calc_vacancy_energy(self, structure, elements):
        if structure not in ['fcc', 'hcp', 'bcc']:
            raise ValueError(f"Cannot calculate vacancy energy for '{structure}'")

        if self.debug:
            print("Calculating vacancy energy...")
            print(40 * "*")
            print(f"Structure: {structure}")
            print(f"Elements: {elements}\n")

        atoms = self.get_structure(structure, elements)
        atoms.set_calculator(self.get_calc())
        e = VacancyEnergy(atoms)

        self.values[('vacancy_energy', structure, elements)] = e

        if self.debug:
            print(f"Vacancy energy: {e:.5f} eV/atom")
            print(40 * "-" + "\n")

    def calc_phase_energy(self, structure, elements):
        if structure not in ['fcc-hcp', 'fcc-bcc', 'bcc-hcp']:
            raise ValueError(f"Cannot calculate phase energy for '{structure}'")

        if self.debug:
            print("Calculating phase energy...")
            print(40 * "*")
            print(f"Structure: {structure}")
            print(f"Elements: {elements}\n")

        if structure[:3] == 'fcc':
            e1 = self.get_cohesive_energy('fcc', elements)
        else:
            e1 = self.get_cohesive_energy('bcc', elements)

        if structure[-3:] == 'hcp':
            e2 = self.get_cohesive_energy('hcp', elements)
        else:
            e2 = self.get_cohesive_energy('bcc', elements)

        self.values[('phase_energy', structure, elements)] = e1 - e2

        if self.debug:
            print(f"Phase energy: {e1 - e2:.5f} eV/atom")
            print(40 * "-" + "\n")

    def calc_elastic_constants(self, structure, elements):
        if structure not in ['fcc', 'hcp', 'bcc', 'l12', 'l10c', 'B2']:
            raise ValueError(f"Cannot calculate elastic constants for '{structure}'")

        if self.debug:
            print("Calculating elastic constants...")
            print(40 * "*")
            print(f"Structure: {structure}")
            print(f"Elements: {elements}\n")

        atoms = self.get_structure(structure, elements)
        atoms.set_calculator(self.get_calc())

        if structure in ['fcc', 'bcc', 'l12', 'B2']:
            C11, C12, C44, B = ElasticConstants(atoms, 'cubic')
            self.values[('elastic_constant_C11', structure, elements)] = C11 / GPa
            self.values[('elastic_constant_C12', structure, elements)] = C12 / GPa
            self.values[('elastic_constant_C44', structure, elements)] = C44 / GPa
            self.values[('bulk_modulus', structure, elements)] = B / GPa
        elif structure == 'l10c':
            B = ElasticConstants(atoms, 'B')
            self.values[('bulk_modulus', structure, elements)] = B / GPa
        else:
            C11, C12, C13, C33, C44, B = ElasticConstants(atoms, 'hexagonal', sanity=False)
            self.values[('elastic_constant_C11', structure, elements)] = C11 / GPa
            self.values[('elastic_constant_C12', structure, elements)] = C12 / GPa
            self.values[('elastic_constant_C13', structure, elements)] = C13 / GPa
            self.values[('elastic_constant_C33', structure, elements)] = C33 / GPa
            self.values[('elastic_constant_C44', structure, elements)] = C44 / GPa
            self.values[('bulk_modulus', structure, elements)] = B / GPa

        if self.debug:
            print(f"C11: {C11 / GPa:.5f} GPa\nC12: {C12 / GPa:.5f} GPa")
            if structure == 'hcp':
                print(f"C13: {C13 / GPa:.5f} GPa\nC33: {C33 / GPa:.5f} GPa")
            print(f"C44: {C44 / GPa:.5f} GPa\nB: {B / GPa:.5f} GPa")
            print(40 * "-" + "\n")

    def calc_elastic_anisotropy(self, structure, elements):
        if structure in ['fcc', 'bcc', 'l12', 'B2']:
            C11 = self.get_elastic_constant_C11(structure, elements)
            C12 = self.get_elastic_constant_C12(structure, elements)
            C44 = self.get_elastic_constant_C44(structure, elements)
            if abs(C11-C12) < 1e-5:
                raise ValueError('Cannot calculate anisotropy: C11=%g  C12=%g  C44=%g'
                                  % (C11, C12, C44))
            self.values[('elastic_anisotropy', structure, elements)] = 2*C44/(C11 - C12)
        else:
            raise ValueError(f"Cannot calculate elastic anisotropy for '{structure}'")

    def calc_lattice_constants(self, structure, elements):
        if structure not in ['fcc', 'hcp', 'bcc', 'l12', 'l10', 'l10c', 'B2', 'A5Bhex']:
            raise ValueError(f"Cannot calculate lattice constants for '{structure}'")

        if self.debug:
            print("Calculating lattice constants...")
            print(40 * "*")
            print(f"Structure: {structure}")
            print(f"Elements: {elements}")

        if structure in ['fcc', 'bcc', 'l12', 'l10c', 'B2']:
            a = self.latticeconstants[(structure, elements)]
            c = None
        else:
            lc = self.latticeconstants[(structure, elements)]
            a = lc[0]
            c = lc[1]

        if self.debug:
            print(f"Guess: {a}, {c}\n")

        atoms = self.get_structure(structure, elements, a=a, c=c)
        atoms.set_calculator(self.get_calc())
        #filter = StrainFilter(atoms)
        #dyn = BFGS(filter, logfile=None, trajectory=None)
        #dyn.run(fmax=0.1)

        if structure in ['fcc', 'bcc', 'l12', 'l10c', 'B2']:
            MinimizeUnitCell(atoms, 'cubic', 0.0001)
            a = atoms.get_cell()[0][0] / self.properties[structure + '_size'][0]
            self.values[('lattice_constant_a', structure, elements)] = a
            #self.latticeconstants[(structure, elements)] = a
        else:
            MinimizeUnitCell(atoms, 'hexagonal', 0.0001)
            a = atoms.get_cell()[0][0] / self.properties[structure + '_size'][0]
            c = atoms.get_cell()[2][2] / self.properties[structure + '_size'][2]
            self.values[('lattice_constant_a', structure, elements)] = a
            self.values[('lattice_constant_c', structure, elements)] = c
            #self.latticeconstants[(structure, elements)] = [a, c]

        E = atoms.get_potential_energy() / len(atoms)
        self.values[('energy', structure, elements)] = E

        if self.debug:
            print(f"a: {a:.5f} A")
            if structure in ['hcp', 'l10']:
                print(f"c: {c:.5f} A")
            print(f"E: {E:.5f} eV")
            print(40 * "-" + "\n")

    def calc_lattice_ratio_ca(self, structure, elements):
        a = self.get_lattice_constant_a(structure, elements)
        c = self.get_lattice_constant_c(structure, elements)

        self.values[('lattice_ratio_ca', structure, elements)] = c / a

    def calc_volume_per_atom(self, structure, elements):
        if structure == 'fcc':
            a = self.get_lattice_constant_a(structure, elements)

            self.values[('volume_per_atom', structure, elements)] = a**3 / 4.0
        elif structure == 'hcp':
            a = self.get_lattice_constant_a(structure, elements)
            c = self.get_lattice_constant_c(structure, elements)

            self.values[('volume_per_atom', structure, elements)] = np.sqrt(3) * a**2 * c / 4.0
        else:
            raise ValueError(f"Cannot calculate lattice constants for '{structure}'")

    def calc_surface_energy(self, structure, elements):
        if structure not in ['fcc100', 'fcc111', 'hcp0001', 'l12100',
                             'l12111']:#, 'hcp1010A', 'hcp1010B']:
            raise ValueError(f"Cannot calculate surface energy for '{structure}'")

        if self.debug:
            print("Calculating surface energy...")
            print(40 * "*")
            print(f"Structure: {structure}")
            print(f"Elements: {elements}\n")

        images = []
        for n in self.properties['surface_layers']:
            images.append(self.get_structure(structure, elements, l=n))

        natoms = len(images[-1]) / self.properties['surface_layers'][-1]
        e = SurfaceEnergy(images, natoms, self.get_calc(), fmax=0.01, debug=self.debug)

        self.values[('surface_energy', structure, elements)] = e

        if self.debug:
            print(f"\nSurface energy: {e:.5f} eV/atom")
            print(40 * "-" + "\n")

    def calc_surface_ratio(self, structure, elements):
        structures = structure.split('-')
        if len(structures) != 2:
            raise ValueError("You must specify exactly two surfaces")

        e = []
        for struct in structures:
            e.append(self.get_surface_energy(struct, elements))

        self.values[('surface_ratio', structure, elements)] = e[0] / e[1]

    def calc_heat_of_formation(self, structure, elements):
        if structure not in ['l12', 'l10', 'l10c', 'B2', 'A5Bhex']:
            raise ValueError(f"Cannot calculate heat of formation for '{structure}'")

        if self.debug:
            print("Calculating heat of formation...")
            print(40 * "*")
            print(f"Structure: {structure}")
            print(f"Elements: {elements}\n")

        e_single = {}
        for symbol in elements:
            sym = reference_states[atomic_numbers[symbol]]['symmetry']
            #atoms = self.get_structure(sym, (symbol,))
            #atoms.set_calculator(self.get_calc())
            #e_single[symbol] = atoms.get_potential_energy() / len(atoms)
            e_single[symbol] = self.get_value('energy', sym, (symbol,),
                                              self.calc_lattice_constants)


        #atoms = self.get_structure(structure, elements)
        #atoms.set_calculator(self.get_calc())
        #e_alloy = atoms.get_potential_energy() / len(atoms)
        e_alloy = self.get_value('energy', structure, elements,
                                 self.calc_lattice_constants)

        if structure == 'l12':
            e = e_alloy - 0.25 * e_single[elements[0]] \
                        - 0.75 * e_single[elements[1]]
        elif structure in ['l10', 'l10c', 'B2']:
            e = e_alloy - 0.5 * e_single[elements[0]] \
                        - 0.5 * e_single[elements[1]]
        elif structure == 'A5Bhex':
            e = e_alloy - 5./6. * e_single[elements[0]] \
                        - 1./6. * e_single[elements[1]]
         
        self.values[('heat_of_formation', structure, elements)] = e

        if self.debug:
            print(f"Heat of formation: {e:.5f} eV/atom")
            print(40 * "-" + "\n")

    #def calc_heat_of_solution(self, structure, elements):
    # lim(N->inf) {E(Pt_N Y) - N*E(Pt) - E(Y)}

    def calc_impurity_energy(self, structure, elements):
        if structure not in ['oct38-center', 'oct38-face', 'oct38-edge']:
            raise ValueError(f"Cannot calculate impurity energy for '{structure}'")

        if len(elements) != 2:
            raise ValueError("Tuple of elements must be of length two")

        if self.debug:
            print("Calculating impurity energy...")
            print(40 * "*")
            print(f"Structure: {structure}")
            print(f"Elements: {elements}\n")

        name, impurity = structure.split('-')
        sym = reference_states[atomic_numbers[elements[1]]]['symmetry']
        latticeconstant = self.get_lattice_constant_a(sym, (elements[1],))

        if name == 'oct38':
            sites = {'center': 10, 'face': 9, 'edge': 0}

            atoms = Octahedron(elements[1], 4, 1, latticeconstant)
            atoms.set_calculator(self.get_calc())

            dyn = BFGS(atoms, logfile=None, trajectory=None)
            dyn.run(fmax=0.001, steps=100)
            assert dyn.converged()
            s_clean = dyn.get_number_of_steps()
            e_clean = atoms.get_potential_energy()

            atoms[sites[impurity]].symbol = elements[0]

            dyn.run(fmax=0.001, steps=100)
            assert dyn.converged()
            s_impurity = dyn.get_number_of_steps()
            e_impurity = atoms.get_potential_energy()

        self.values[('impurity_energy', structure, elements)] = e_impurity - e_clean

        if self.debug:
            print("BFGS steps: %i    %i" % (s_clean, s_impurity))
            print("Impurity energy: {:.5f} - {:.5f} = {:.5f} eV".format(e_impurity, e_clean,
                                                              e_impurity - e_clean))
            print(40 * "-" + "\n")

    def calc_impurity_ratio(self, structure, elements):
        structures = structure.split('/')
        if len(structures) != 2:
            raise ValueError("You must specify exactly two structures")

        e = []
        for struct in structures:
            e.append(self.get_impurity_energy(struct, elements))

        self.values[('impurity_ratio', structure, elements)] = e[0] / e[1]

    def calc_cutoff_energy(self, structure, elements):
        if structure not in ['fcc', 'hcp', 'l12']:
            raise ValueError(f"Cannot calculate cutoff energy for '{structure}'")

        if self.debug:
            print("Calculating cutoff energy...")
            print(40 * "*")
            print(f"Structure: {structure}")
            print(f"Elements: {elements}\n")

        if structure in ['fcc', 'l12']:
            a = self.latticeconstants[(structure, elements)]
            c = None
            f = 1.73
        else:
            lc = self.latticeconstants[(structure, elements)]
            a = lc[0]
            c = lc[1]
            f = 1.73

        atoms = self.get_structure(structure, elements, a=a, c=c)
        atoms.set_calculator(self.get_calc())

        e_equilibrium = atoms.get_potential_energy()

        new_cell = f * atoms.get_cell()
        atoms.set_cell(new_cell, True)

        e_cutoff = atoms.get_potential_energy()

        self.values[('cutoff_energy', structure, elements)] = e_cutoff / e_equilibrium

        if self.debug:
            print(f"Equilibrium: {e_equilibrium:.5f} eV")
            print(f"Cutoff: {e_cutoff:.5f} eV")
            print(f"Ratio: {e_cutoff / e_equilibrium:.5f}")
            print(40 * "-" + "\n")

    def calc_scaling_energy(self, structure, elements):
        t = structure.split('-')
        structure = t[0]
        scaling = float(t[1])

        if structure not in ['fcc', 'hcp', 'l12']:
            raise ValueError(f"Cannot calculate scaling energy for '{structure}'")

        if self.debug:
            print("Calculating scaling energy...")
            print(40 * "*")
            print(f"Structure: {structure}")
            print(f"Elements: {elements}\n")

        if structure in ['fcc', 'l12']:
            a = self.latticeconstants[(structure, elements)]
            c = None
            f = 1.73
        else:
            lc = self.latticeconstants[(structure, elements)]
            a = lc[0]
            c = lc[1]
            f = 1.73

        atoms = self.get_structure(structure, elements, a=a, c=c)
        atoms.set_calculator(self.get_calc())
        new_cell = scaling * atoms.get_cell()
        atoms.set_cell(new_cell, True)
        energy = atoms.get_potential_energy() / len(atoms)

        structure = f'{structure}-{scaling:.2f}'
        self.values[('scaling_energy', structure, elements)] = energy

        if self.debug:
            print(f"Energy: {energy:.5f} eV")
            print(40 * "-" + "\n")

    def calc_force_match(self, structure, elements):
        if self.debug:
            print("Calculating force matching...")
            print(40 * "*")
            print(f"Structure: {structure}")
            print(f"Elements: {elements}\n")
        atoms = Trajectory(structure)[-1]
        f_dft = atoms.get_forces()
        atoms = atoms.repeat((2,2,2))
        atoms.set_calculator(self.get_calc())
        f = atoms.get_forces()[:len(f_dft)]
        df = f_dft - f
        err = np.sqrt((df*df).sum())
        err /= np.sqrt((f_dft*f_dft).sum())
        self.values[('force_match', structure, elements)] = err
        if self.debug:
            print(f"Mean force error: {err:.5f} eV/A")
            print(40 * "-" + "\n")


    def calc_stacking_fault(self, structure, elements):
        if structure != 'fcc':
            raise ValueError("Cannot calculate stacking fault energy of structure "+structure)
        a = self.latticeconstants[(structure, elements)]
        atoms = fcc111(elements[0], (1,2,5), orthogonal=True, a=a, vacuum=a/(2 * np.sqrt(3.0)))
        atoms.set_pbc(True)
        atoms = atoms.repeat(self.properties['sf_repeat'])
        atoms.set_calculator(self.get_calc())
        dyn = QuasiNewton(atoms, logfile=None, trajectory=None)
        dyn.run(fmax=0.02)
        e_sf = atoms.get_potential_energy()
        n_sf = len(atoms)
        atoms = fcc111(elements[0], (1,2,6), orthogonal=True, a=a, vacuum=a/(2 * np.sqrt(3.0)))
        atoms.set_pbc(True)
        atoms = atoms.repeat(self.properties['sf_repeat'])
        atoms.set_calculator(self.get_calc())
        dyn = QuasiNewton(atoms, logfile=None, trajectory=None)
        dyn.run(fmax=0.02)
        e_bulk = atoms.get_potential_energy()
        n_bulk = len(atoms)
        layers = self.properties['sf_repeat'][2]
        uc = atoms.get_cell()
        area = uc[0,0] * uc[1,1]
        result = (e_sf - e_bulk * n_sf / n_bulk) / layers / area
        result /= 1e-3 * kJ * 1e-20
        self.values[('stacking_fault', structure, elements)] = result
        
    ### Bulk structures ###
    def get_structure(self, name, elements, a=None, c= None, l=None):
        # Check number of elements
        if name[:3] in ['fcc', 'hcp']:
            if len(elements) != 1:
                raise ValueError("Tuple of elements must be of length one")
        if name[:3] in ['l12', 'l10'] or name[:2] == 'B2' or name[:6] == 'A5Bhex':
            if len(elements) != 2:
                raise ValueError("Tuple of elements must be of length two")

        # Get lattice constants
        if a is None:
            if name[:2] == 'B2':
                a = self.get_lattice_constant_a(name[:2], elements)
            elif name[:4] in ['l10c']:
                a = self.get_lattice_constant_a(name[:4], elements)
            elif name[:3] in ['fcc', 'hcp', 'bcc', 'l12', 'l10', 'l10c']:
                a = self.get_lattice_constant_a(name[:3], elements)

        if c is None:
            if name[:3] in ['hcp', 'l10'] and name[:4] != 'l10c':
                c = self.get_lattice_constant_c(name[:3], elements)

        if a < 0.5:
            raise ValueError("Cannot create structure with a < 0.5")
        if c is not None and c < 0.5:
            raise ValueError("Cannot create structure with c < 0.5")
        
        # Get size
        if name in ['fcc100', 'fcc111', 'hcp0001']:
            size = self.properties[name + '_size'][:2] + (l,)
        else:
            size = self.properties[name + '_size']
            

        # Make structure
        if name == 'fcc':
            atoms = FaceCenteredCubic(symbol=elements[0],
                                      size=size,
                                      latticeconstant=a)
        elif name == 'hcp':
            atoms = HexagonalClosedPacked(symbol=elements[0], size=size,
                                          directions=[[2,-1,-1,0],[0,1,-1,0],[0,0,0,1]],
                                          latticeconstant=(a, c))
        elif name == 'bcc':
            atoms = BodyCenteredCubic(symbol=elements[0],
                                      size=size,
                                      latticeconstant=a)
        elif name == 'B2':
            atoms = B2(symbol=elements, size=size, latticeconstant=a)
        elif name == 'l12':
            atoms = L1_2(symbol=elements, size=size, latticeconstant=a)
        elif name == 'l10':
            atoms = L1_0(symbol=elements, size=size, latticeconstant=(a, c))
        elif name == 'l10c':
            atoms = L1_0(symbol=elements, size=size, latticeconstant=(a, a))
        elif name == 'A5Bhex':
             atoms = crystal((elements[1], elements[0], elements[0]),
                             basis=[(0.0, 0.0, 0.0),
                                    (1.0/3.0, 2.0/3.0, 0.0), 
                                    (0.5, 0.0, 0.5)],
                             spacegroup=191, 
                             cellpar=[a, a, c, 90, 90, 120])
        elif name == 'fcc100':
            atoms = fcc100(symbol=elements[0], size=size, a=a, vacuum=10.0)
        elif name == 'fcc111':
            atoms = fcc111(symbol=elements[0], size=size, a=a,
                           vacuum=10.0, orthogonal=True)
        elif name == 'hcp0001':
            atoms = hcp0001(symbol=elements[0], size=size, a=a, c=c,
                            vacuum=10.0, orthogonal=True)
        elif name == 'hcp1010A':
            raise ValueError(f"Structure '{name}' not supported")
            atoms = None
        elif name == 'hcp1010B':
            raise ValueError(f"Structure '{name}' not supported")
            atoms = None
        elif name == 'l12100':
            n = (l + 1) / 2
            atoms = L1_2(symbol=elements, size=(8, 8, n), latticeconstant=a)
            atoms.set_pbc([True, True, False])
            # Remove layers
            atoms = atoms[atoms.get_positions()[:,2] > 0.1 * a]
            # Set vacuum
            atoms.center(axis=2, vacuum=10.0)
        elif name == 'l12111':
            if l % 3 == 0:
                n = l / 3
                c = 0
            else:
                n = l / 3 + 1
                c = 3 - l % 3
            atoms = L1_2(symbol=elements, size=(8, 4, n),
                         #directions=[[1,-1,0],[1,0,-1],[1,1,1]], latticeconstant=a)
                         directions=[[1,-1,0],[1,1,-2],[1,1,1]], latticeconstant=a)
            atoms.set_pbc([True, True, False])
            # Wrap positions
            scpos = atoms.get_scaled_positions(wrap=True)
            scpos[scpos > (1.0 - 1e-12)] = 0.0
            atoms.set_scaled_positions(scpos)
            # Remove layers
            if c > 0:
                atoms = atoms[atoms.get_positions()[:,2] > (c - 0.5) * a / np.sqrt(3.0)]
            # Set vacuum
            atoms.center(axis=2, vacuum=10.0)
        else:
            raise ValueError(f"Structure '{name}' not supported")
        return atoms
    
    def get_calc(self):
        if callable(self.calc):
            return self.calc()
        else:
            return self.calc

def MinimizeUnitCell(atoms, symmetry, ftol):
    from asap3.Tools.ParameterOptimization.ScipyFmin import fmin
    if symmetry == 'cubic':
        var = [1.0]  # Scaling
    elif symmetry == 'hexagonal':
        var = [1.0, 1.0]
    else:
        raise ValueError(f'Symmetry "{symmetry}" not supported')
    u0 = atoms.get_cell()
    def energy(v, a=atoms, u=u0):
        newcell = v[0] * u
        if len(v) > 1:
            newcell[2] *= v[1]
        if newcell[0,0] < 0.5 or newcell[1,1] < 0.5 or newcell[2,2] < 0.5:
            print("ERROR: Absurd unit cell")
            print(newcell)
            print("Material: ", atoms.get_chemical_formula())
            raise ValueError("Absurd unit cell: " + str(newcell.flatten()))
        a.set_cell(newcell, scale_atoms=True)
        return a.get_potential_energy()
    xopt, fopt, iter, calls, flag = fmin(energy, var, ftol=ftol, delta=0.01,
                                         full_output=True, disp=False)
    newcell = xopt[0] * u0
    if len(xopt) > 1:
        newcell[2] *= xopt[1]
    atoms.set_cell(newcell, scale_atoms=True)

if __name__ == '__main__':
    from asap3 import EMT, EMT2013
    from asap3.Tools.Timing import Timer
    from asap3.Tools.ParameterOptimization.EMT import EMT2011Fit

    timer = Timer()
    quantities = [('lattice_constant_a', 'fcc', 'Pt', 3.92),
                  ('lattice_constant_a', 'hcp', 'Pt', 2.77),
                  ('lattice_ratio_ca', 'hcp', 'Pt', 4.78 / 2.77),
                  ('bulk_modulus', 'fcc', 'Pt', 278.3),
                  ('elastic_anisotropy', 'fcc', 'Pt', 1.594),
                  #('elastic_constant_C11', 'fcc', 'Pt', 346.7),
                  #('elastic_constant_C12', 'fcc', 'Pt', 250.7),
                  ('elastic_constant_C44', 'fcc', 'Pt', 76.5),
                  ('cohesive_energy', 'fcc', 'Pt', 5.84),
                  #('phase_energy', 'fcc-hcp', 'Pt', -0.05),
                  ('surface_energy', 'fcc111', 'Pt', 0.631),
                  ('surface_ratio', 'fcc111-fcc100', 'Pt', 0.631 / 0.892),
                  #('lattice_constant_a', 'l12', ('Pt', 'Cu'), 3.67860),
                  #('heat_of_formation', 'l12', ('Pt', 'Cu'), -0.12331),
                  ]

    latticeconstants = [('fcc', 'Pt', 3.92),
                        ('hcp', 'Pt', (2.77, 4.78)),
                        ('fcc', 'Cu', 3.5),
                        ('l12', ('Pt', 'Cu'), 3.8)]

    
    calc = QuantityCalc(latticeconstants, EMT(), timer, True)
    for name, structure, elements, value in quantities:
        value = calc.get_quantity(name, structure, elements)

    timer.write()




#! /usr/bin/env python

from numpy import *
from asap3 import *
from asap3.md.verlet_fast import VelocityVerlet
from asap3.md.langevin import Langevin
from ase.lattice.cubic import FaceCenteredCubic
from asap3.timing import report_timing
import sys, pickle, time, commands, os, re
import numpy as np
from asap3.testtools import ReportTest

# cpu time:  time.clock().   Wall clock time: time.time()

#Verbose(1)

usethread = (len(sys.argv) > 1 and sys.argv[1] == "-t")
if usethread:
    AsapThreads()


host = commands.getoutput("hostname")
timesteps = 100
selfcheckfilename = "timing-selfcheck.dat"
asapversion = get_version()
when = time.strftime("%a %d %b %Y %H:%M", time.localtime(time.time()))

randomstate = "randomstate.pickle"
if os.path.isfile(randomstate):
    np.random.set_state(pickle.load(open(randomstate, "rb")))
else:
    print("Saving random state for next call.")
    rndfile = open(randomstate, "wb")
    pickle.dump(np.random.get_state(), rndfile)
    rndfile.close()
    

#PrintVersion(1)
print("Running ASAP timing on "+host+".")
if re.match(r"^n\d\d\d.dcsc.fysik.dtu.dk$", host):
    print("    This is a d512 node on Niflheim.")
    fullhost = "niflheim-d512/%s" % (host.split(".")[0])
    host = "niflheim-d512"
elif re.match(r"^[stu]\d\d\d.dcsc.fysik.dtu.dk$", host):
    print("    This is an s50 node on Niflheim.")
    fullhost = "niflheim-s50/%s" % (host.split(".")[0])
    host = "niflheim-s50"
else:
    fullhost = host
print("Current time is "+when)
print("")

print("Preparing system")
initial = FaceCenteredCubic(directions=[[1,0,0],[0,1,0],[0,0,1]],
                            size=(30, 30, 30),
                            symbol="Cu")
ReportTest("Number of atoms", len(initial), 108000, 0)
r = initial.get_positions()
r.flat[:] += 0.14 * sin(arange(3*len(initial)))
initial.set_positions(r)

print("Running self-test.")
atoms = Atoms(initial)
atoms.set_calculator(EMT())
e = atoms.get_potential_energies()
f = atoms.get_forces()
if os.access(selfcheckfilename, os.F_OK):
    olde, oldf = pickle.load(open(selfcheckfilename, "rb"))
    de = max(fabs(e - olde))
    df = max(fabs(f.flat[:] - oldf.flat[:]))
    print("Maximal deviation:  Energy", de, "  Force", df)
    ReportTest("Max force error", df, 0.0, 1e-11)
    ReportTest("Max energy error", de, 0.0, 1e-11)
    del olde, oldf
else:
    print("WARNING: No self-check database found, creating it.")
    pickle.dump((e, f), open(selfcheckfilename, "wb"))
del e,f,atoms

ReportTest.Summary(exit=1)

print("Preparing to run Verlet dynamics.")
atoms = Atoms(initial)
atoms.set_calculator(EMT())
dynamics = VelocityVerlet(atoms, 5*units.fs)

print("Running Verlet dynamics.")
startcpu, startwall = time.clock(), time.time()
dynamics.run(timesteps)

vcpu, vwall = time.clock() - startcpu, time.time() - startwall
vfraction = vcpu/vwall
sys.stderr.write("\n")
print("Verlet dynamics done.")
del dynamics, atoms

print("")
print("")
print("TIMING RESULTS:")
print(f"Verlet:   CPU time {vcpu:.2f}s  Wall clock time {vwall:.2f}s ({vfraction * 100:.0f}%)")
print("")

report_timing()



"""Python access to NeighborLocators."""

from asap3 import _asap
import numpy as np

def _decide_minimumimage(rCut, atoms, driftfactor=0.0):
    # Auto-detect minimum-image conditions
    if atoms is None:
        # No information, assuming minimum image conditions.
        minimum_image = True
    else:
        cell = atoms.cell
        heights = cell.volume / cell.areas()
        toosmall = np.less_equal(heights, rCut * (1 + driftfactor))
        toosmall = np.logical_and(toosmall, atoms.get_pbc())
        minimum_image = not toosmall.any()
    return minimum_image

def FullNeighborList(rCut, atoms=None, driftfactor=0.0, minimum_image=None):
    if minimum_image is None:
        minimum_image = _decide_minimumimage(rCut, atoms)
    return _asap.FullNeighborList(rCut, atoms, driftfactor, minimum_image)




from asap3 import Atoms, EMT, EMTRasmussenParameters, EMT2013
from asap3.md.verlet import VelocityVerlet
from asap3.EMT2013Parameters import sihb_PtY_parameters
from ase.lattice.cubic import FaceCenteredCubic
from ase.lattice.compounds import L1_2
from ase.data import atomic_masses_legacy
from asap3.test.pytest_markers import serial

import numpy as np

import os
import json

import pytest

pytestmark = [pytest.mark.core, serial]

@pytest.fixture
def potResults(datadir):
    with open(datadir / 'potResults.json', 'rt') as resfile:
        return json.load(resfile)
    
@pytest.fixture
def copper():
    return FaceCenteredCubic(directions=[[1,0,0],[0,1,0],[0,0,1]], 
                             size=(15,15,15), symbol="Cu",
                             pbc=(1,0,1), debug=0)

@pytest.fixture
def cu_au_3():
    return L1_2(directions=[[1,0,0],[0,1,0],[0,0,1]], size=(15,15,15),
                symbol=("Cu", "Au"), latticeconstant=3.95, pbc=(1,0,1), 
                debug=0)

@pytest.fixture
def platinum():
    return FaceCenteredCubic(directions=[[1,0,0],[0,1,0],[0,0,1]],
                             size=(15,15,15), symbol="Pt", pbc=(1,0,1), 
                             debug=0)

@pytest.fixture
def yttrium():
    return FaceCenteredCubic(directions=[[1,0,0],[0,1,0],[0,0,1]], 
                             size=(15,15,15), symbol="Y", 
                             latticeconstant=4.97,
                             pbc=(1,0,1), debug=0)

@pytest.fixture
def pt_3_y():
    return L1_2(directions=[[1,0,0],[0,1,0],[0,0,1]], size=(15,15,15),
                symbol=("Y", "Pt"), latticeconstant=4.06, pbc=(1,0,1), 
                debug=0)

def test_EMT_Cu(copper, potResults):
    atoms = Atoms(copper)
    atoms.calc = EMT()
    dotest(atoms, 50, 0.1, "EMT_Cu", potResults)

def test_EMT_Cu_Rasm(copper, potResults):
    atoms = Atoms(copper)
    atoms.calc = EMT(EMTRasmussenParameters())
    dotest(atoms, 50, 0.1, "EMT_Cu_Rasm", potResults)

def test_EMT_CuAu3(cu_au_3, potResults):
    atoms = Atoms(cu_au_3)
    nCu = np.sum(np.equal(atoms.get_atomic_numbers(), 29))
    nAu = np.sum(np.equal(atoms.get_atomic_numbers(), 79))
    assert nCu == 13500 / 4, 'Number of Cu atoms in alloy'
    assert nAu == 3 * 13500 / 4, 'Number of Au atoms in alloy'
    atoms.calc = EMT()
    dotest(atoms, 50, 0.06, "EMT_CuAu3", potResults)

def test_EMT2013_Pt(platinum, potResults):
    atoms = Atoms(platinum)
    atoms.calc = EMT2013(sihb_PtY_parameters)
    dotest(atoms, 50, 0.1, "EMT2013_Pt", potResults)

def test_EMT2013_Y(yttrium, potResults):
    atoms = Atoms(yttrium)
    atoms.calc = EMT2013(sihb_PtY_parameters)
    dotest(atoms, 50, 0.1, "EMT2013_Y", potResults)

def test_EMT2013_Pt3Y(pt_3_y, potResults):
    atoms = Atoms(pt_3_y)
    atoms.calc = EMT2013(sihb_PtY_parameters)
    dotest(atoms, 50, 0.06, "EMT2013_Pt3Y", potResults)

def dotest(atoms, nsteps, ampl, name, potResults):
    assert len(atoms) == 13500, f'Number of atoms ({name})'
    timeunit = 1.018047e-14             # Seconds
    femtosecond = 1e-15 / timeunit      # Femtosecond in atomic units
    atoms.set_masses(atomic_masses_legacy[atoms.numbers])
    print("Potential energy", atoms.get_potential_energy() / len(atoms))
    r = atoms.get_positions()
    r.flat[:] += ampl * np.sin(np.arange(3*len(atoms)))
    atoms.set_positions(r)
    print("Potential energy", atoms.get_potential_energy() / len(atoms))

    print(f"Running Verlet dynamics ({name})")
    dyn = VelocityVerlet(atoms, 2*femtosecond)
    etot1 = (atoms.get_potential_energy() + atoms.get_kinetic_energy())
    dyn.run(nsteps)
    etot2 = (atoms.get_potential_energy() + atoms.get_kinetic_energy())
    assert np.isclose(etot1, etot2, atol=1.0), f"Energy conservation ({name})"
    print(etot1, etot2)

    epot = atoms.get_potential_energies()
    stress = atoms.get_stresses(include_ideal_gas=True)

    print("Testing energies and stresses")
    j = 0
    eres=potResults["e"+name]
    sres=potResults["s"+name]
    for i in range(len(atoms)//100):
        assert np.isclose(epot[i*100], eres[i], atol=1e-8), f'{name} energy {i*100}'
        assert np.isclose(stress[i*100, j], sres[i], atol=1e-8), f'{name} stress {i*100}'
        j = (j + 1) % 6



import asap3
from asap3.mpi import world
import numpy as np
from asap3.test.pytest_markers import ReportTest, parallel

import pytest

@pytest.mark.core
@parallel
def test_simplempi(cpulayout):
    "Simple test of Asap's MPI object."
    outdata = np.array([world.rank * 2], int)
    indata = np.zeros(1, int)

    to = world.rank + 1
    if to >= world.size:
        to = 0
    fr = world.rank - 1
    if fr < 0:
        fr += world.size

    print("%d: to=%d from=%d" % (world.rank, to, fr))

    rq = world.send(outdata, to, block=False)
    world.receive(indata, fr)
    world.wait(rq)

    ReportTest(f"Data received by {world.rank} from {fr}", indata[0], 2*fr, 0)


    outdata = world.rank * np.arange(10000000)
    indata = np.zeros(10000000, int)
    print("Data types:", outdata.shape, outdata.dtype, indata.shape, indata.dtype)

    to = world.rank + 1
    if to >= world.size:
        to = 0
    fr = world.rank - 1
    if fr < 0:
        fr += world.size

    print("%d: to=%d from=%d" % (world.rank, to, fr))

    rq = world.send(outdata, to, block=False)
    print("%d: Nonblocking send returned" % (world.rank,))
    print("%d: %s" % (world.rank, repr(rq)))
    world.receive(indata, fr)
    print("%d: Receive returned" % (world.rank,)) 
    #world.wait(rq)
    print("Request completed:", rq.test())
    rq.wait()

    assert (indata == fr * np.arange(10000000)).all()

    s = world.sum(2.5 * world.rank)
    ReportTest("Float sum", s, 2.5 * 0.5 * world.size * (world.size - 1), 1e-9)

    s = world.sum(world.rank)
    ReportTest("Integer sum", s, world.size * (world.size - 1) / 2, 0)

    s = world.max(1.0 * world.rank)
    ReportTest("Float max", s, world.size - 1, 0)

    s = world.max(world.rank)
    ReportTest("Int max", s, world.size - 1, 0)

    print("Checking reduces on arrays...")
    d0 = np.arange(10) + world.rank
    f0 = d0 + 0.0

    d = d0.copy()
    f = f0.copy()
    world.max(d)
    assert (d == np.arange(10) + world.size - 1).all()
    world.max(f)
    assert (f == np.arange(10) + world.size - 1.0).all()

    d = d0.copy()
    f = f0.copy()
    world.min(d)
    assert (d == np.arange(10)).all()
    world.min(f)
    assert (f == np.arange(10)).all()

    print("  ... passed.")

    print("Checking new communicator")
    newcomm = world.new_communicator(np.arange(2))
    if newcomm:
        print("I am", newcomm.rank)
        s = newcomm.sum(world.rank+1)
        ReportTest("Integer sum", s, 3, 0)

    world.barrier()



from asap3 import EMT, EMT2013, OpenKIMsupported
if OpenKIMsupported:
    from asap3 import OpenKIMcalculator
from asap3.EMT2013Parameters import sihb_PtY_parameters
from ase.lattice.cubic import FaceCenteredCubic
from ase.lattice.compounds import L1_2
from ase.data import reference_states, atomic_numbers
from asap3.test.pytest_markers import ReportTest, serial, _openkimmodel
import numpy as np

import pytest


# Using fixtures is possible but inconvenient when parametrizing test,
# and since nothing is really gained we just use normal functions.
def make_L1_2(symbol, size, pbc):
    lc = [reference_states[atomic_numbers[s]]['a'] for s in symbol]
    lc = (3*lc[0] + lc[1])/4.0
    return L1_2(symbol=symbol, size=size, latticeconstant=lc, pbc=pbc)

testtypes = [
    # element, structure, pot, potargs, lefthanded, name
    ('Au', FaceCenteredCubic, EMT, (), False, 'EMT'),
    (('Au', 'Ag'), make_L1_2, EMT, (), False, 'EMT(alloy)'),
    ('Au', FaceCenteredCubic, EMT, (), True, 'EMT(lefthanded)'),
    ('Pt', FaceCenteredCubic, EMT2013, (sihb_PtY_parameters,), False, 'EMT2013'),
    (('Y', 'Pt'), make_L1_2, EMT2013, (sihb_PtY_parameters,), False, 'EMT2013(alloy)')
]
if OpenKIMsupported:
    testtypes += [
        ('Au', FaceCenteredCubic, OpenKIMcalculator, (_openkimmodel,), False, 'OpenKIM'),
    ]

@serial
@pytest.mark.parametrize('element,makestructure,pot,potargs,lefthanded,_', 
                         testtypes, ids=[x[-1] for x in testtypes])
def test_size_consistent(element, makestructure, pot, potargs, lefthanded, _):
    sizes = np.arange(10,0,-1)

    energy = None
    force = None
    for s in sizes:
        atoms = makestructure(symbol=element, size=(s,s,s), pbc=True)
        if lefthanded:
            uc = atoms.get_cell()
            uc[2] *= -1.0
            atoms.set_cell(uc, scale_atoms=True)
        r = atoms.get_positions()
        for i in range(0, len(r), 4):
            r[i][0] += 0.1
        atoms.set_positions(r)
        atoms.calc = pot(*potargs)
        e = atoms.get_potential_energy() / len(atoms)   
        print("%5i atoms: E = %.5f eV/atom" % (len(atoms), e))
        if energy is None:
            energy = e
        else:
            ReportTest("Energy for size %i (%i atoms)" % (s, len(atoms)),
                    e, energy, 1e-8)
        f = atoms.get_forces()
        if force is None:
            force = f[:4]
        else:
            for i in range(len(f)):
                for j in range(3):
                    ReportTest("Force for size %i atom %i of %i component %i"
                            % (s, i, len(atoms), j),
                            f[i,j], force[i % 4, j], 1e-8)
        del atoms


import numpy as np
# Delayed import:  scipy.optimize
# It is imported in the minimization function, so the main functionality
# of this module is preserved even if SciPy is not installed.

def ElasticConstants(atoms, symmetry, minimize=False, **kwargs):
    """Calculate the anisotropic elastic constants of a system.
    
    The atoms must be a periodic system with an orthorhombic cell,
    and the lattice constant(s) should be the equilibrium value(s).  
    If not, the function will crash, unless called with 
    minimize=True to perform an (inefficient) minimization first.
    
    The atoms are modified, but returned to the original state, 
    except if minimize=True in which case the unit cell is optimized.
    
    The symmetry of the system must be specified.  Currently, 
    the following symmetries are allowed::
    
    symmetry='cubic'
        Returns C11, C12, C44 and B.
        Note that B = (C11 + 2*C12)/3
        
    symmetry='hexagonal'
        Returns C11, C12, C13, C33, C44 and B.
        Note that B is given by the other constants (but the
        expression is not trivial)

    symmetry="B"
        Returns B
        
    Note that the system must reflect the underlying symmetry 
    (not checked), and that the unit cell must be orthorombic 
    (may be relaxed later for hexagonal systems).
    """   
    ucell = atoms.get_cell()
    for i in range(3):
        for j in range(3):
            if i != j and abs(ucell[i,j]) > 1e-10:
                raise ValueError("Unit cell of atoms is not orthorhombic.")     
    if minimize:
        minimize_unit_cell(atoms, symmetry)
    if symmetry.lower() == 'cubic':
        return _elastic_constants_cubic(atoms, **kwargs)
    elif symmetry.lower() == 'b':
        return _elastic_constants_bulkonly(atoms, **kwargs)
    elif symmetry.lower() == 'hexagonal':
        return _elastic_constants_hexagonal(atoms, **kwargs)
    else:
        raise ValueError(f'Symmetry "{symmetry}" not supported')
    
def _elastic_constants_cubic(atoms, debug=False):
    """Calculate C11, C12, C44 and B for a system with cubic symmetry."""
    e0 = atoms.get_potential_energy()
    u0 = atoms.get_cell()
    C11 = elastic_constant('C11', atoms, e0, u0, (1,0,0,0,0,0),
                           debug=debug)
    B = elastic_constant('B', atoms, e0, u0, 
                         (1/3.0, 1/3.0, 1/3.0, 0, 0, 0), 
                         debug=debug)
    C44 = elastic_constant('C44', atoms, e0, u0, 
                           (0,0,0,1/2.0,0,0), debug=debug)
    C12 = (3*B - C11)/2.0
    return (C11, C12, C44, B)

def _elastic_constants_bulkonly(atoms, debug=False):
    """Calculate B for a system with cubic symmetry."""
    e0 = atoms.get_potential_energy()
    u0 = atoms.get_cell()
    B = elastic_constant('B', atoms, e0, u0, 
                         (1/3.0, 1/3.0, 1/3.0, 0, 0, 0), 
                         debug=debug)
    return B

def _elastic_constants_hexagonal(atoms, debug=False, sanity=True):
    """Calculate C11, C12, C13, C33, C44 and B for a hexagonal system."""
    e0 = atoms.get_potential_energy()
    u0 = atoms.get_cell()
    C11 = elastic_constant('C11', atoms, e0, u0, (1,0,0,0,0,0),
                           debug=debug)
    C33 = elastic_constant('C33', atoms, e0, u0, (0,0,1,0,0,0),
                           debug=debug)
    C11_C12 = elastic_constant('2C11+2C12', atoms, e0, u0,
                               (1,1,0,0,0,0), debug=debug) / 2
    C12 = C11_C12 - C11
    C11_C33_2C13 = elastic_constant('C11+C33+2C13', atoms, e0, u0,
                                    (1,0,1,0,0,0), debug=debug)
    C13 = 0.5 * (C11_C33_2C13 - C11 - C33)
    C44 = elastic_constant('C44', atoms, e0, u0, 
                           (0,0,0,1/2.0,0,0), debug=debug)
    C66 = elastic_constant('C66', atoms, e0, u0, 
                           (0,0,0,0,0,1/2.0), debug=debug)
    C66_alt = 0.5 * (C11 - C12)
    devC66 = 2 * abs(C66 - C66_alt) / (C66 + C66_alt) 
    if debug or devC66 > 0.02:
        print(f"Deviation in C66: {100 * devC66:.1f}%")
    if devC66 > 0.02 and sanity:
        raise RuntimeError("Deviation in C66: %.1f%% > 2%%: Not a hexagonal system?"
                           % (100 * devC66,))
    alpha = (C11 + C12 - 2*C13) / (C33 - C13)
    if debug:
        print("alpha =", alpha)
    B = (2 * C11 + 2 * C12 + 4 * alpha * C13 
         + C33 * alpha * alpha) / ((2 + alpha) * (2 + alpha))
    return (C11, C12, C13, C33, C44, B)

def elastic_constant(name, atoms, e0, u0, mode, strain=0.007, debug=False):
    """Calculate an elastic constant.
    
    Parameters::
    
    name
        Name of the constant (a string). 
        
    atoms
        The atoms.
        
    e0
        Energy in the equilibrium configuration.
        
    u0
        Unit cell in the equilibrium configuration.
        
    mode
        The deformation mode as six numbers, giving weights
        to the six strain components.
    """
    strainlist = (-1.0, 0.5, 0, 0.5, 1.0)
    energies = []
    order = np.array([[0, 5, 4],
                      [5, 1, 3],
                      [4, 3, 2]])
    straintensor = np.array(mode)[order]
    if debug:
        print(f"{name} unit strain tensor:")
        print(straintensor)
    for s in strainlist:
        if s == 0:
            energies.append(e0)
        else:
            # Apply strain
            s = s * strain * straintensor + np.identity(3)
            atoms.set_cell(np.dot(u0, s), scale_atoms=True)
            energies.append(atoms.get_potential_energy())
    atoms.set_cell(u0, scale_atoms=True)  # Reset atoms
    fit0 = np.poly1d(np.polyfit(strain * np.array(strainlist), energies, 3))
    fit1 = np.polyder(fit0, 1)
    fit2 = np.polyder(fit1, 1)
    x0 = None
    for x in np.roots(fit1):
        if fit2(x) > 0:
            if x0 is not None:
                raise RuntimeError("More than two roots found.")
            assert x0 is None
            x0 = x
    if x0 is None:
        raise RuntimeError("No roots found.")
    if np.abs(x0) > 0.5 * strain:
        raise RuntimeError(f"Unreasonable root ({x0:f}): " +
                           "Maybe the system was not at the equilibrium configuration")
    if debug:
        print("Root:", x0)
    value = fit2(x0)
    return value / atoms.get_volume()

def minimize_unit_cell(atoms, symmetry):
    #from scipy.optimize import fmin_powell, fmin
    from asap3.Tools.ParameterOptimization.ScipyFmin import fmin
    if symmetry == 'cubic':
        var = [1.0]  # Scaling
    elif symmetry == 'hexagonal':
        var = [1.0, 1.0]
    else:
        raise ValueError(f'Symmetry "{symmetry}" not supported')
    u0 = atoms.get_cell()
    def energy(v, a=atoms, u=u0):
        newcell = v[0] * u
        if len(v) > 1:
            newcell[2] *= v[1]
        a.set_cell(newcell, scale_atoms=True)
        #if len(v) > 1:
        #    print "fmin: %.4f  %7.4f  %8.4f  %8.4f" % (v[0], v[1], a.get_positions()[1792,2], a.get_potential_energy())
        return a.get_potential_energy()
    xopt, fopt, iter, calls, flag = fmin(energy, var, delta=0.01, full_output=True, disp=False)
    #print "Minimize unit cell:", xopt, fopt, iter, calls

if __name__ == '__main__':
    from asap3 import EMT, units, EMThcpParameters
    from ase.lattice.cubic import FaceCenteredCubic
    from ase.lattice.hexagonal import HexagonalClosedPacked
    
    print("Calculating cubic constants for Cu")
    atoms = FaceCenteredCubic(size=(5,5,5), symbol='Cu', 
                              latticeconstant=3.59495722231)
    atoms.set_calculator(EMT())
    e = ElasticConstants(atoms, 'cubic')
    print(np.array(e) / units.GPa)
    
    print("Pretending that Cu is hexagonal")
    e = ElasticConstants(atoms, 'hexagonal', sanity=False)
    print(np.array(e) / units.GPa)
    
    print("Calculating elastic constants for Ru")
    atoms = HexagonalClosedPacked(size=(5,5,5), symbol='Ru',
                                  directions=[[1,-2,1,0], 
                                              [1,0,-1,0],
                                              [0,0,0,1]])
    atoms.set_calculator(EMT(EMThcpParameters()))
    e = ElasticConstants(atoms, 'hexagonal', minimize=True)
    print(np.array(e) / units.GPa)
    print("The EMT values are not even close to experimental values!")
    
    print("Pretending Ru is cubic")
    e = ElasticConstants(atoms, 'cubic')
    print(np.array(e) / units.GPa)
    



from asap3 import *
from asap3.constraints import FixAtoms
from ase.lattice.cubic import FaceCenteredCubic
from ase.io import read, write
from asap3.io import Trajectory
from asap3.test.pytest_markers import ReportTest, serial

import os
import numpy as np

import pytest

usemask = True

@serial
@pytest.mark.parametrize('usemask', (True, False))
def test_constr_in_traj(usemask, in_tmp_dir):
    fn = "testconstraint"
    atoms = FaceCenteredCubic(symbol='Cu', size=(3,3,3))
    if usemask:
        c = FixAtoms(mask=np.arange(len(atoms))<3)
    else:
        c = FixAtoms(indices=(2,3))
    atoms.set_constraint(c)
    print(c)

    traj = Trajectory(fn+'.trj', "w", atoms)
    traj.write()
    traj.close()
    del traj

    traj = BundleTrajectory(fn+'.bundle', "w", atoms)
    traj.write()
    traj.close()
    del traj

    traj = Trajectory(fn+'.traj', "w", atoms)
    traj.write()
    traj.close()
    del traj

    traj2 = Trajectory(fn+'.trj')
    atoms2 = traj2[-1]
    c2 = atoms2.constraints[0]

    print("*** Trajectory")
    print("Original class:", c.__class__)
    print("New class:", c2.__class__)
    assert c.__class__ is c2.__class__

    traj2 = read(fn+'.bundle')
    c2 = atoms2.constraints[0]

    print("*** BundleTrajectory")
    print("Original class:", c.__class__)
    print("New class:", c2.__class__)
    assert c.__class__ is c2.__class__

    print("*** Cleaning up")
    os.unlink(fn+'.traj')
    os.unlink(fn+'.trj')
    BundleTrajectory.delete_bundle(fn+'.bundle')

    print("Test passed.")




from ase.md.nptberendsen import NPTBerendsen as _NPTBerendsen
from ase.md.nptberendsen import Inhomogeneous_NPTBerendsen as _Inhomogeneous_NPTBerendsen
from asap3.md.md import ParallelMolDynMixinNoData

class NPTBerendsen(_NPTBerendsen, ParallelMolDynMixinNoData):
    def __init__(self, atoms, timestep, temperature=None,
                 *, fixcm=False, **kwargs):
        """Berendsen (constant N, P, T) molecular dynamics.

        This dynamics scale the velocities and volumes to maintain a constant
        pressure and temperature.  The shape of the simulation cell is not
        altered, if that is desired use Inhomogenous_NPTBerendsen.

        Parameters:

        atoms: Atoms object
            The list of atoms.

        timestep: float
            The time step in ASE time units.

        temperature: float
            The desired temperature, in Kelvin.

        temperature_K: float
            Alias for ``temperature``.

        pressure: float (deprecated)
            The desired pressure, in bar (1 bar = 1e5 Pa).  Deprecated,
            use ``pressure_au`` instead.

        pressure: float
            The desired pressure, in atomic units (eV/Å^3).

        taut: float
            Time constant for Berendsen temperature coupling in ASE
            time units.  Default: 0.5 ps.

        taup: float
            Time constant for Berendsen pressure coupling.  Default: 1 ps.

        compressibility: float (deprecated)
            The compressibility of the material, in bar-1.  Deprecated,
            use ``compressibility_au`` instead.

        compressibility_au: float
            The compressibility of the material, in atomic units (Å^3/eV).

        fixcm: bool (optional)
            If True, the position and momentum of the center of mass is
            kept unperturbed.  Default: False.   This is not supported in
            parallel simulations, and should in general be unnecessary unless
            a global non-zero center-of-mass momentum needs to be preserved,
            as a zero center-of-mass momentum will remain zero.

        trajectory: Trajectory object or str (optional)
            Attach trajectory object.  If *trajectory* is a string a
            Trajectory will be constructed.  Use *None* for no
            trajectory.

        logfile: file object or str (optional)
            If *logfile* is a string, a file with that name will be opened.
            Use '-' for stdout.

        loginterval: int (optional)
            Only write a log line for every *loginterval* time steps.  
            Default: 1

        append_trajectory: boolean (optional)
            Defaults to False, which causes the trajectory file to be
            overwriten each time the dynamics is restarted from scratch.
            If True, the new structures are appended to the trajectory
            file instead.

        """
        if (fixcm and getattr(atoms, 'parallel', False)):
            raise NotImplementedError("NVTBerendsen does not support fixcm=True for parallel simulations.")
        super().__init__(atoms, timestep, temperature, fixcm=fixcm, **kwargs)

    def run(self, steps):
        self.before_run()
        super().run(steps)
        self.after_run()


class Inhomogeneous_NPTBerendsen(_Inhomogeneous_NPTBerendsen, ParallelMolDynMixinNoData):
    def __init__(self, atoms, timestep, temperature=None,
                 *, fixcm=False, **kwargs):
        if (fixcm and getattr(atoms, 'parallel', False)):
            raise NotImplementedError("NVTBerendsen does not support fixcm=True for parallel simulations.")
        super().__init__(atoms, timestep, temperature, fixcm=fixcm, **kwargs)

    def run(self, steps):
        self.before_run()
        super().run(steps)
        self.after_run()



"""
Tests that Asap works in parallel simulations even if a process has no atoms.

Name: emptyNode.py

Description: Part of the Asap test suite.  Tests parallelization robustness.

Usage: mpirun -np 4 asap-python emptyNode.py

Expected result: The output should end with 'ALL TESTS SUCCEEDED'.
"""

from ase.build import bulk
from ase.lattice.cubic import FaceCenteredCubic
from asap3 import MakeParallelAtoms, EMT
from asap3.mpi import world
import numpy as np
from ase.visualize import view

from asap3.test.pytest_markers import ReportTest, parallel

import pytest

def makesystem(delta):
    atoms = FaceCenteredCubic("Cu", size=(15,15,15), pbc=False)
    x = atoms.cell[0,0] / 2
    y = atoms.cell[1,1] / 2
    pos = atoms.get_positions()
    keep = np.logical_or(pos[:,0] > x + delta, pos[:,1] > y + delta)
    return atoms[keep]

def distribute(atoms, cpulayout):
    if world.rank != 0:
        atoms = None
    return MakeParallelAtoms(atoms, cpulayout)

def checksystem(origatoms, label, cpulayout):
    origatoms.calc = None
    atoms = distribute(origatoms, cpulayout)
    atoms.suppress_warning_noatoms = True   # Kill a warning on stderr
    print(f"Number of atoms on process {world.rank}: {len(atoms)}")
    atoms.calc = EMT()
    energy_par = atoms.get_potential_energy()

    origatoms.calc = EMT()
    energy_ser = origatoms.get_potential_energy()

    ReportTest(label+": Energy match", energy_par, energy_ser, 1e-6)

@pytest.mark.skipif(world.size != 4, reason='Requires 4 MPI tasks')
def test_empty_nodes(cpulayout):
    a = makesystem(10.0)
    checksystem(a, "Large missing part on master", cpulayout)

    a.rotate(90, 'z', center='COU')
    checksystem(a, "Large missing part on slave", cpulayout)

    a = makesystem(1.0)
    checksystem(a, "Small missing part on master", cpulayout)

    a.rotate(90, 'z', center='COU')
    checksystem(a, "Small missing part on slave", cpulayout)



#!/usr/bin/env python

from asap3 import Atoms, EMT
from ase.lattice.cubic import FaceCenteredCubic
from asap3.setup.dislocation import Dislocation
from asap3.analysis import CNA
# from asap3.visualize.primiplotter import *
from asap3.test.pytest_markers import ReportTest, serial
import numpy as np

import pytest


@serial
def test_make_dislocation():
    splitting = 5
    #size = (50, 88, 35)
    size = (30, 25, 7)

    Gold = "Au"
    slab = FaceCenteredCubic(directions=((1,1,-2), (-1,1,0), (1,1,1)),
                            size=size, symbol=Gold)
    basis = slab.get_cell()
    print(basis)
    print("Number of atoms:", len(slab))

    center = 0.5 * np.array([basis[0,0], basis[1,1], basis[2,2]]) + np.array([0.1, 0.1, 0.1])
    offset = 0.5 * splitting * slab.miller_to_direction((-1,0,1))
    print(center)

    d1 = Dislocation(center - offset, slab.miller_to_direction((-1,-1,0)),
                    slab.miller_to_direction((-2,-1,1))/6.0)
    d2 = Dislocation(center + offset, slab.miller_to_direction((1,1,0)),
                    slab.miller_to_direction((1,2,1))/6.0)

    atoms = Atoms(slab)
    (d1+d2).apply_to(atoms)
    del slab

    print("Now running CNA")

    atoms.calc = EMT()
    c = CNA(atoms)
    atoms.set_tags(c)

    expected = [27703, 314, 3483]

    for i in range(3):
        print(i, sum(np.equal(c, i)))
        ReportTest("Number of atoms with CNA class "+str(i), sum(np.equal(c, i)), expected[i], 0)




"""Defines functions and observers analysing the local structure.

The following FUNCTIONS are defined.  They operate on the atoms, and
return an array with the resulting data.  They do not modify the
atoms, but may trigger neighbor list updates in the attached
potential.

CNA: Simple CNA analysis, finds atoms with local FCC and HCP
coordination. 

CoordinationNumbers: Finds the number of neighbors of each atom.

The following CLASSES are defined.  They are can be used as
observers (or base classes for observers) of the atoms by attaching 
their update method to the dynamics.  They can also have their own
observers, so fx a trajectory can be told to write *after* the analysis
is done.

RestrictedCNA: Simple CNA analysis, finds atoms with local FCC and HCP
coordination.  The tags of the atoms are set accordingly.

FullCNA: Full CNA analysis.  Several methods are available for returning
different kind of CNA information.  To use as an observer, the update
method MUST be overridden and augmented to call one of the get_xxx_cna
methods, and use the returned information somehow.

"""

import numpy as np
from asap3 import _asap
from asap3.mpi import world
import ase.data
from asap3.Internal.ListOfElements import ListOfElements
from asap3.Internal.Subject import Subject
from asap3.Internal.NeighborLocators import _decide_minimumimage

def GuessLatticeConstant(atoms):
    """Gets an estimate of the lattice constant for a system.

    If only a single element is present, the lattice constant for that
    element is returned.

    If more than one element is present, a weighted average of their
    lattice constants is returned.

    In any case, the lattice constant is the 'a' constant if the
    crystal structure is not cubic, except in the case of HCP structure
    where sqrt(2)*a is returned.

    Note that this function is only intended for guessing a cutoff
    radius for e.g. CNA analysis, do not rely on this lattice constant
    being very meaningful.

    If any element does not have a lattice constant, a KeyError is raised.
    """
    refstate = ase.data.reference_states
    loe = ListOfElements(atoms)
    if len(loe) == 1:
        if len(atoms):
            z = atoms.get_atomic_numbers()
            assert(z.min() == z.max())
            # Cast to int is a numpy bug workaround: np.int32 objects are not ints
            if refstate[int(z[0])]['symmetry'] == 'hcp':
                latconst = refstate[int(z[0])]['a'] * np.sqrt(2)
            else:
                latconst = refstate[int(z[0])]['a']
        else:
            latconst = 0.0
    else:
        if len(atoms):
            a = 0.0
            n = 0
            z = atoms.get_atomic_numbers()
            for e in loe:
                n1 = np.equal(z, e).sum() # Number of atoms with this element
                if refstate[int(e)]['symmetry'] == 'hcp':
                    a += n1 * refstate[int(e)]['a'] * np.sqrt(2)
                else:
                    a += n1 * refstate[int(e)]['a']
                n += n1
            latconst = a / n
        else:
            latconst = 0.0
    if world.size > 1:
        latconst = world.sum(latconst * len(atoms)) / atoms.get_global_number_of_atoms()
    return latconst

def _checkcell(r, atoms, name):
    pbc = atoms.get_pbc()
    if pbc.any():
        cell = atoms.cell
        heights = cell.volume / cell.areas()
        minheight = heights[pbc].min()
        if minheight < r:
            raise RuntimeError(f'{name}: The height of the cell {minheight:.3f} '
                               + f'must be larger than {r}.')
    
def CNA(atoms, rCut=None):
    """Restricted Common Neighbor Analysis on a list of atoms.

    Returns an array of integers (int8) reflecting the local crystal structure.
    FCC is class 0, HCP is class 1 and everything else is class 2.

    It requires a value of rCut between the first and second shell.
    It tries to guess a value for rCut from the atomic numbers if none
    is specified.

    Consider using `Asap.Filters.RestrictedCNA` instead.
    """
    if not rCut:
        rCut = GuessLatticeConstant(atoms) * 0.825
    _checkcell(2 * rCut, atoms, 'CNA')
    return _asap.RestrictedCNA(atoms, rCut)


def CoordinationNumbers(atoms, rCut = None, minimum_image=None):
    """Coordination numbers of the atoms.

    Returns the coordination numbers of the atoms.  The cutoff can be
    given, otherwise it tries to guess a value between the first and
    second coordination shell, based on the atomic numbers.
    """
    if not rCut:
        rCut = GuessLatticeConstant(atoms) * 0.825
    if minimum_image is None:
        minimum_image = _decide_minimumimage(rCut, atoms)
    return _asap.CoordinationNumbers(atoms, rCut, minimum_image)

class FullCNA:
    """Full Common Neighbor Analysis on a list of atoms
    
    This CNA calculator can be used as a stand-alone analyzer or
    as an observer.  To calculate the CNA of some atoms once, 
    create the object and call the appropriate get_cna function.
        
    The same object can be used to calculate the CNA multiple times
    when an atoms object is altered (although little if anything is
    saved by reusing the object).  Each time a get_cna_xxx() method
    is called, the CNA are recalculated.  If you use a diffent
    atoms object or want to change the cutoff, call the update
    method.
    
    To use as an observer, derive a subclass and give it a method
    which gets the relevant CNA data and stores it (or whatever 
    the observer should do with the data).
    """
    def __init__(self, atoms=None, rCut=None):
        self.atoms = atoms
        self.rawcna = None
        self.rCut = rCut
        if self.rCut is None and atoms is not None:
            self.rCut = GuessLatticeConstant(self.atoms) * 0.825
        if atoms is not None:
            _checkcell(2 * self.rCut, self.atoms, 'FullCNA')
        self.tuples = {}
    
    def update(self, atoms=None, rCut=None):
        """Change the atoms or the cutoff."""
        if atoms is not None:
            self.atoms = atoms
        if self.atoms is None:
            raise ValueError("Atoms must be specified when creating the object or when calling update.")
        if rCut is not None:
            self.rCut = rCut
        if not self.rCut:
            self.rCut = GuessLatticeConstant(self.atoms) * 0.825
        _checkcell(2 * self.rCut, self.atoms, 'FullCNA')

    def get_number_of_pairs(self):
        return len(self.get_raw_cna())

    def get_raw_cna(self):
        """Return raw CNA data as a numpy array."""
        return _asap.FullCNA(self.atoms, self.rCut).get_raw_cna()
    
    def get_cna_tuple(self, cnaint):
        """Convert a CNA integer to a CNA tuple."""
        try:
            return self.tuples[cnaint]
        except KeyError:
            cna = ((cnaint & 0xFF0000) >> 16, (cnaint & 0xFF00) >> 8, cnaint & 0xFF)
            self.tuples[cnaint] = cna
            return cna
    
    def get_normal_cna(self):
        """CNA as a dictionary per atom.
        
        The keys of the dictionaries are the CNA tuples, the values are 
        the number of times they occur.
        """
        return _asap.FullCNA(self.atoms, self.rCut).get_per_atom_cna()
        # The code below implements the same in Python.  It is too slow but is
        # kept for inspiration
#        rawcna = self.get_raw_cna()
#        n = len(self.atoms)
#        result = [{} for i in range(n)]
#        for bond in rawcna:
#            for atom in bond[:2]:
#                assert(atom < n)
#                cna = self.get_cna_tuple(bond[2])
#                try:
#                    result[atom][cna] += 1
#                except KeyError:
#                    result[atom][cna] = 1
#        return result
    
    def get_total_cna(self):
        """Get the total CNA for the system.
        
        A dictionary is returned with the CNA tuples as keys, and the total
        number of bonds with that CNA index as values.
        """
        return _asap.FullCNA(self.atoms, self.rCut).get_total_cna()
        # The code below implements the same in Python.  It is slower but is
        # kept for inspiration
#        rawcna = self.get_raw_cna()
#        temp = {}
#        for x in rawcna[:,2]:
#            try:
#                temp[x] += 1
#            except KeyError:
#                temp[x] = 1
#        result = {}
#        for k, v in temp.iteritems():
#            result[self.get_cna_tuple(k)] = v
#        return result
    
    def get_total_perz_cna(self):
        """Get the total CNA for the system per element.
        
        A dictionary is returned where the keys are the pairs of
        elements encountered in the system, and the keys are
        dictionaries summarizing the CNA (as for get_total_cna())
        for bonds between these elements.  The pairs of elements
        are always given with the smallest atomic number first.
        """

        rawcna = self.get_raw_cna()
        z = self.atoms.get_atomic_numbers()
        temp = {}
        rawcna[:,:2] = z[rawcna[:,:2]]
        # Summarize the data with minimal processing
        for line in rawcna:
            try:
                temp[tuple(line)] += 1
            except KeyError:
                temp[tuple(line)] = 1
        # Now do the processing.
        # First, sum data with z1 > z2 into entries with z1 < z2
        for z1, z2, cna in temp.keys():
            if z2 < z1:
                temp[(z2, z1, cna)] = temp.get((z2, z1, cna), 0) + temp[(z1,z2,cna)]
        result = {}
        for k in temp.keys():
            z1, z2, cna = k
            if (z1 <= z2):
                # Valid entry
                try:
                    d = result[(z1, z2)]
                except KeyError:
                    d = result[(z1, z2)] = {}
                d[self.get_cna_tuple(cna)] = temp[k]
        return result
    
    def get_normal_and_total_cna(self):
        """Return the per-atom and total cna.
        
        Returns a tuple with two elements, the first is
        the same as get_normal_cna() would return, the second
        is what get_total_cna() would return.  However, unlike
        calling the two methods separately, the actual CNA 
        calculation is only done once.
        """
        # This is intended as an example of how to return
        # the CNA in multiple ways without repeating the calculations
        workhorse = _asap.FullCNA(self.atoms, self.rCut)
        normal = workhorse.get_per_atom_cna()
        total = workhorse.get_total_cna()
        return (normal, total)
    
class RestrictedCNA(Subject):
    """Restricted Common Neighbor analysis on a list of atoms.

    Sets the tags of the atoms according to their local crystal
    structure.  FCC is class 0, HCP is class 1 and everything else is
    class 2.
    
    It requires a value of rCut between the first and second shell.
    It tries to guess a value for rCut from the atomic numbers if none
    is specified.
    
    This class is intended to use as an observer, so its analyze()
    function is called automatically by e.g. the dynamics.  It can
    itself act as a subject, so a Plotter or a Trajectory can be
    called just after the calculations.

    Optional parameters:
        
        rCut = None:  The cutoff.  If not given, one is guessed.

        analyze_first = True: Should an analysis be made as soon as
        this object is created?  Leave as true if you plan to attach a
        Trajectory as an observer to this object, and if the
        Trajectory will save the initial state.
        
    """
    def __init__(self, atoms, rCut=None, analyze_first=True):
        Subject.__init__(self)
        self.atoms = atoms
        if rCut:
            self.rCut = rCut
        else:
            self.rCut = GuessLatticeConstant(atoms) * 0.825
        _checkcell(2 * self.rCut, self.atoms, 'RestrictedCNA')
        if analyze_first:
            self.analyze()  # There will not be any observers yet.

    def analyze(self):
        "Runs the CNA analysis."
        cna = CNA(self.atoms, self.rCut)
        self.atoms.set_tags(cna)
        self.call_observers()
        
    update = analyze
    



import numpy as np

from ase import units
from ase.data import atomic_numbers
from ase.md import VelocityVerlet
from ase.md.velocitydistribution import MaxwellBoltzmannDistribution
from ase.lattice.cubic import SimpleCubic
from asap3 import Morse, set_verbose
from asap3.testtools import ReportTest
import time

np.random.seed(42)

#set_verbose(2)

# Define constants and calculator
elements = np.array([atomic_numbers['Ru'], atomic_numbers['Ar']])
epsilon = np.array([[5.720, 0.092], [0.092, 0.008]])
alpha = np.array([[1.475, 2.719], [2.719, 1.472]])
rmin = np.array([[2.110, 2.563], [2.563, 4.185]])

calc = Morse(elements, epsilon, alpha, rmin)

# Define atoms
atoms = SimpleCubic('Ar', size=(15,15,15), latticeconstant=50.0)
n = 0
while n < 100:
    i = np.random.randint(len(atoms)-1)
    if atoms[i].number != atomic_numbers['Ru']:
        atoms[i].number = atomic_numbers['Ru']
        n += 1
atoms.set_calculator(calc)

# Set initial momentum
MaxwellBoltzmannDistribution(atoms, 1000*units.kB)

# Run dynamics
startcpu, startwall = time.clock(), time.time()

dyn = VelocityVerlet(atoms, 1.0 * units.fs, logfile='test-energy.dat', loginterval=10)
dyn.run(10)
etot = (atoms.get_potential_energy() + atoms.get_kinetic_energy())/len(atoms)
#obs.close()
for i in range(25):
    if i:
        dyn.run(1000)
    epot = atoms.get_potential_energy()/len(atoms)
    ekin = atoms.get_kinetic_energy()/len(atoms)
    print(f"{epot:9.5f} {ekin:9.5f}   {epot+ekin:9.5f}")
    ReportTest("Step %i." % (i,), epot+ekin, etot, 1e-3, silent=True)

cpu, wall = time.clock() - startcpu, time.time() - startwall
fraction = cpu/wall

print("")
print("")
print("TIMING RESULTS:")
print(f"Morse potential:   CPU time {cpu:.2f}s  Wall clock time {wall:.2f}s ({fraction*100:.0f}%)")
print("")

ReportTest.Summary()



import numpy as np
from ase import units
import sys
sys.path.append('.')
from AdsorptionParameters import *
#Changes: Added dictionary to include multiple objects.
    
def getEnergies(species):
    
    p = adsorptionparameters(species)
    
    return p.get('Eads') 
    
    #return np.array([-0.25,-0.25,-0.25,-0.25,-0.25,-0.25,-0.2,-0.2,0.041,-0.107
    #,100,100,100,100,100,100])


def getCoverages(T,P,species):
    #This script calculates the expression for the Langmuir isotherm(a function of Temperature and pressure):
    #It returns coverages ordered by coordinations such that CN=0 ~ index 0
    
    p = adsorptionparameters(species)
    
    #Constants:
    Pstd=1E5 #1E5 Pa is standard pressure
    Tstd = 298.15 #25 cecluis degrees
    
    Viben = np.array(p.get('viben')) #Vibration Energies
    Sgas = p.get('Sgas') #Entropy of gas (standard conditions)
    dissociative = p.get('dissoc')  # True if dissociative, false if not.

    cp = 7./2*units.kB #Per atom diving by 2.heat capacity of diatomic ideal gas for constant pressure.
   
    #Adsorption Energies:
    dE = p.get('Eads')

    #Calculate entropy of vibrations:
    Svib = units.kB*((Viben/units.kB/T/(np.exp(Viben/units.kB/T)-1)-np.log(1-np.exp(-Viben/units.kB/T))).sum())
    
    #Rate Constants:
    dF = dE-T*(Svib-(Sgas+cp*np.log(T/Tstd)-units.kB*np.log(P/Pstd))/2.0)
    K = np.exp(-dF/units.kB/T)
    #Coverage Langmuir dissociative adsorption:
    if dissociative:
        theta = np.sqrt(K)/(1+np.sqrt(K))
    else:
        theta = K/(1+K)
    
    return theta


